# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
doge_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/doge'
    },
    {
        'url': 'https://dogecoin-api.flare.network/'
    },
    {
        'url': 'https://dogechain.rpc.thirdweb.com/'
    },
    {
        'url': 'https://rpc.dogechain.dog'
    }
]


import requests
import hashlib

from ...errors import NetworkException

async def broadcast_transaction_eth_blockcypher(raw_transaction_hex):
    api_url = "https://api.blockcypher.com/v1/eth/main/txs/push"
    payload = {"tx": raw_transaction_hex}

    try:
        response = requests.post(api_url, json=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))

    if response.status_code == 201:
        return hashlib.sha256(hashlib.sha256(raw_transaction_hex.encode()).digest()).digest()  # Transaction ID
    else:
        raise NetworkException("Failed to broadcast Ethereum transaction using BlockCypher API: {}".format(response.text))



import os
import binascii
import hashlib
from typing import AnyStr, Union, List


string_types = (str)
string_or_bytes_types = (str, bytes)
int_types = (int, float)
# Base switching
code_strings = {
    2: '01',
    10: '0123456789',
    16: '0123456789abcdef',
    32: 'abcdefghijklmnopqrstuvwxyz234567',
    58: '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz',
    256: ''.join([chr(x) for x in range(256)])
}

def bin_dbl_sha256(s):
    bytes_to_hash = from_string_to_bytes(s)
    return hashlib.sha256(hashlib.sha256(bytes_to_hash).digest()).digest()

def lpad(msg, symbol, length):
    if len(msg) >= length:
        return msg
    return symbol * (length - len(msg)) + msg

def get_code_string(base):
    if base in code_strings:
        return code_strings[base]
    else:
        raise ValueError("Invalid base!")


def changebase(string, frm, to, minlen=0):
    if frm == to:
        return lpad(string, get_code_string(frm)[0], minlen)
    return encode(decode(string, frm), to, minlen)


def bin_to_b58check(inp: bytes, magicbyte: int = 0) -> str:
    if magicbyte == 0:
        inp = from_int_to_byte(0) + inp
    while magicbyte > 0:
        inp = from_int_to_byte(magicbyte % 256) + inp
        magicbyte //= 256

    leadingzbytes = 0
    for x in inp:
        if x != 0:
            break
        leadingzbytes += 1

    checksum = bin_dbl_sha256(inp)[:4]
    return '1' * leadingzbytes + changebase(inp+checksum, 256, 58)


def bytes_to_hex_string(b: Union[int, bytes, List[str], List[int]]) -> str:
    if isinstance(b, str):
        return b

    return ''.join('{:02x}'.format(y) for y in b)


def safe_from_hex(s: str) -> bytes:
    return bytes.fromhex(s)


def from_int_representation_to_bytes(a: int) -> bytes:
    return bytes(str(a), 'utf-8')


def from_int_to_byte(a: int) -> bytes:
    return bytes([a])


def from_byte_to_int(a: bytes) -> int:
    return a


def from_string_to_bytes(a: AnyStr) -> bytes:
    return a if isinstance(a, bytes) else bytes(a, 'utf-8')


def safe_hexlify(a: bytes) -> str:
    return str(binascii.hexlify(a), 'utf-8')


def encode(val, base, minlen=0):
    base, minlen = int(base), int(minlen)
    code_string = get_code_string(base)
    result_bytes = bytes()
    while val > 0:
        curcode = code_string[val % base]
        result_bytes = bytes([ord(curcode)]) + result_bytes
        val //= base

    pad_size = minlen - len(result_bytes)

    padding_element = b'\x00' if base == 256 else b'1' \
        if base == 58 else b'0'
    if (pad_size > 0):
        result_bytes = padding_element*pad_size + result_bytes

    result_string = ''.join([chr(y) for y in result_bytes])
    result = result_bytes if base == 256 else result_string

    return result


def decode(string, base):
    if base == 256 and isinstance(string, str):
        string = bytes(bytearray.fromhex(string))
    base = int(base)
    code_string = get_code_string(base)
    result = 0
    if base == 256:
        def extract(d, cs):
            return d
    else:
        def extract(d, cs):
            return cs.find(d if isinstance(d, str) else chr(d))

    if base == 16:
        string = string.lower()
    while len(string) > 0:
        result *= base
        result += extract(string[0], code_string)
        string = string[1:]
    return result


def random_string(x):
    return str(os.urandom(x))


from .esplora import EsploraAddress

class MempoolSpaceAddress(EsploraAddress):
    def __init__(self, addresses, request_interval=(3,1), transactions=None):
        super().__init__(addresses, request_interval=request_interval, transactions=transactions, url="https://mempool.space/testnet/api")


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = [
    'bcy',
    'btc',
    'btctest',
    'dash',
    'dashtest',
    'doge',
    'dogetest',
    'eth',
    'ltc',
    'ltctest'
]

from .broadcast import broadcast_transaction, tx_hash


from .utils.keys import PublicKey
from .utils.base58 import b58decode_check
from .utils.bech32 import bech32_decode
# Amounts are always described internally in the lowest possible denomination
# to make them integers.
class Destination:
    def __init__(self, address, amount, network):
        self._network = network
        self._address = address
        self._amount = amount
        
        self._script_pubkey = PublicKey.script(address, network)
                

    def address(self):
        return self._address
    
    def amount(self, in_standard_units=True):
        if not in_standard_units:
            if self._network.SUPPORTS_EVM:
                return int(self._amount * 1e18)
            else:
                return int(self._amount * 1e8)
        else:
            return self._amount
    
    def script_pubkey(self):
        return self._script_pubkey

# destination has no attribute script_pubkey.... line 228, encode.py


# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
dogetest_nodes = [
    {
        'url': 'https://dogechain-testnet.rpc.thirdweb.com/' #!!!! these are ETHEREUM nodes, not DOGE nodes?
    },
    {
        'url': 'https://rpc-testnet.dogechain.dog' # Who's silly idea was that?
    },
]


import requests
import time
import websocket
import json

from functools import reduce

from ...errors import NetworkException
from ...utils.utils import convert_to_utc_timestamp
from ...generated import wallet_pb2

def deduplicate(elements):
    return reduce(lambda re, x: re+[x] if x not in re else re, elements, [])

class DogeChainAddress:
    """
    A class representing a Dogecoin address.

    This class allows you to retrieve the balance and transaction history of a Dogecoin address using the DogeChain API.

    Args:
        address (str): The human-readable Dogecoin address.

    Attributes:
        address (str): The human-readable Dogecoin address.

    Methods:
        get_balance(): Retrieves the balance of the Dogecoin address.
        get_transaction_history(): Retrieves the transaction history of the Dogecoin address.

    Raises:
        Exception: If the API request fails or the address balance/transaction history cannot be retrieved.
    """

    def _clean_tx(self, element):
        url = f"https://dogechain.info/api/v1/transaction/{element['hash']}"
        for attempt in range(3, -1, -1):
            if attempt == 0:
                raise NetworkException("Network request failure")
            try:
                response = requests.get(url, timeout=60)
                if response.status_code == 200:
                    data = element
                    element = response.json()['transaction']
                    break
                else:
                    raise NetworkException("Failed to retrieve transaction history")
            except requests.RequestException:
                pass

        new_element = wallet_pb2.Transaction()
        new_element.txid = element['hash']
        
        if element['confirmations'] == 0:
            new_element.confirmed = False
            new_element.height = 0
        else:
            new_element.confirmed = True
            url = f"https://dogechain.info/api/v1/block/{element['block_hash']}"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, timeout=60)
                    if response.status_code == 200:
                        new_element.height = response.json()['block']['height']
                        break
                    else:
                        raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass
        
        new_element.timestamp = element['time']
        
        for vin in element['inputs']:
            txinput = new_element.btclike_transaction.inputs.add()
            txinput.txid = '' if 'previous_output' not in vin.keys() else vin['previous_output'].get('hash') or ''
            txinput.index = 0 if 'previous_output' not in vin.keys() else vin['previous_output'].get('pos') or 0
            txinput.amount = 0 if 'value' not in vin.keys() else int(float(vin['value']) * 1e8)
        
        i = 0
        for vout in element['outputs']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(float(vout['value']) * 1e8)
            txoutput.index = i
            i += 1
            txoutput.address = vout['address']
            txoutput.spent = vout['spent'] is not None
        
        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])
        new_element.total_fee = total_inputs - total_outputs

        new_element.btclike_transaction.fee = int(float(element['fee']) * 1e8)
        new_element.fee_metric = wallet_pb2.BYTE
        
        return new_element

    def __init__(self, addresses, request_interval=(3,1), transactions=None):
        """
        Initializes an instance of the DogeChainAddress class.

        Args:
            addresses (list): A list of human-readable Dogecoin addresses.
            request_interval (tuple): A pair of integers indicating the number of requests allowed during
                a particular amount of seconds. Set to (0,N) for no rate limiting, where N>0.
        """
        self.addresses = addresses
        self.requests, self.interval_sec = request_interval
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.transactions = deduplicate([*self._get_transaction_history()])
        self.height = self.get_block_height()

    def get_balance(self):
        """
        Retrieves the balance of the Dogecoin address.

        Returns:
            float: The balance of the Dogecoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos

    def get_block_height(self):
        """Returns the current block height."""

        # Dogechain gives us the block height through a web socket
        # This only works because the block mining for Dogecoin is very fast.
        uri = "wss://ws.dogechain.info/inv"
        for attempt in range(3, -1, -1):
            if attempt == 0:
                                raise NetworkException("Network request failure")
            try:
                ws = websocket.create_connection(uri)
                message = {"op": "ping_block"}
                ws.send(json.dumps(message))

                # Ignore the pong response
                ws.recv()
                # And get the real response
                response = json.loads(ws.recv())
            except requests.RequestException:
                pass
            finally:
                try:
                    ws.close()
                except Exception:
                    pass
                break
        
        self.height = response['x']['height']
        return self.height
        

    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Dogecoin address from cached data augmented with network data.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs

        self.transactions = deduplicate(self.transactions)
        return self.transactions

    def _get_transaction_history(self, txhash=None):
        """
        Retrieves the transaction history of the Dogecoin address. (internal method that makes the network query)

        Parameters:
            txhash (str): Get all transactions before (and not including) txhash.
                Defaults to None, which disables this behavior.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        for address in self.addresses:
            i = 1
            url = f"https://dogechain.info/api/v1/address/transactions/{address}/{i}"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, timeout=60)
                    if response.status_code == 200:
                        data = response.json()
                        break
                    else:
                                                raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass

            for tx in data["transactions"]:
                time.sleep(self.interval_sec/(self.requests*len(data["transactions"])))
                if txhash and tx["hash"] == txhash:
                    return
                yield self._clean_tx(tx)
            if not data["transactions"]:
                return
            
            while data["transactions"]:
                i += 1
                url = f"https://dogechain.info/api/v1/address/transactions/{address}/{i}"
                for attempt in range(3, -1, -1):
                    if attempt == 0:
                                                raise NetworkException("Network request failure")
                    try:
                        response = requests.get(url, timeout=60)
                        break
                    except requests.RequestException:
                        pass

                if response.status_code == 200:
                    data = response.json()
                    for tx in data["transactions"]:
                        time.sleep(self.interval_sec/(self.requests*len(data["transactions"])))
                        if txhash and tx["hash"] == txhash:
                            return
                        yield self._clean_tx(tx)
                    if not data["transactions"]:
                        return
                else:
                                        raise NetworkException("Failed to retrieve transaction history")



import requests

from ...errors import NetworkException

async def broadcast_transaction_btc_esplora(raw_transaction_hex, **kwargs):
    esplora_url = kwargs.get('url')
    api_url = f"{esplora_url}/tx"
    payload = {"tx": raw_transaction_hex}

    try:
        response = requests.post(api_url, data=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))

    if response.status_code >= 300:
        raise NetworkException(f"Failed to broadcast transaction using Esplora API: {response.text}")



"""
This file contains parameters and constants for all supported networks.
"""

from .generated import wallet_pb2

# Internal base class: Do not use.
class CryptoNetwork(object):
    pass

class BitcoinCashMainNet(CryptoNetwork):
    """ Bitcoin Cash MainNet version bytes. """
    NAME = "Bitcoin Cash"
    INTERNAL_NAME = wallet_pb2.Network.BITCOIN_CASH_MAINNET
    COIN = "BCH"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x28  # int(0x28) = 40
    PUBKEY_ADDRESS = 0x1C  # int(0x00) = 28  # Used to create payment addresses
    SECRET_KEY = 0x80      # int(0x80) = 128  # Used for WIF format
    # same as Bitcoin
    EXT_PUBLIC_KEY = 0x0488b21E  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x0488ADE4  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/145'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class DashMainNet(CryptoNetwork):
    """ Dash MainNet version bytes.
    
    OK, time for a little drama - Dash's xpub/xprv were originally drkp/drkv.

    But somehow, they were swapped when inerted into the refernece client.
    So there are actually two possible combinations of addresses in Dash
    (3, if you also count Bitcoin's xpub/xprv)

    See: https://www.dash.org/forum/index.php?threads/dash-bip32-serialization-values-dev-discussion-wont-apply-to-most.8092/
    """
    NAME = "Dash"
    INTERNAL_NAME = wallet_pb2.Network.DASH_MAINNET
    COIN = "DASH"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x10  # int(0x10) = 16
    PUBKEY_ADDRESS = 0x4C  # int(0x4C) = 76  # Used to create payment addresses
    SECRET_KEY = 0xCC      # int(0xCC) = 204  # Used for WIF format
    EXT_PUBLIC_KEY = 0x02fe52cc  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x02fe52f8  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/5'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported

class DashInvertedMainNet(CryptoNetwork):
    """Dash MainNet version bytes.
    
    This is the version that uses drkv/drkp as the public/private
    extended version bytes respectively. It intentionally has the same name
    and coin as the other Dash mainnets.
    """
    NAME = "Dash"
    INTERNAL_NAME = wallet_pb2.Network.DASH_INVERTED_MAINNET
    COIN = "DASH"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x10  # int(0x10) = 16
    PUBKEY_ADDRESS = 0x4C  # int(0x4C) = 76  # Used to create payment addresses
    SECRET_KEY = 0xCC      # int(0xCC) = 204  # Used for WIF format
    EXT_PUBLIC_KEY = 0x02fe52f8  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x02fe52cc  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/5'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported

class DashBTCMainNet(CryptoNetwork):
    """Dash MainNet version bytes.
    
    Extended version bytes are the same as for Bitcoin, i.e. xpub/xprv,
    for maximum wallet compatibility.
    """
    NAME = "Dash"
    INTERNAL_NAME = wallet_pb2.Network.DASH_BTC_MAINNET
    COIN = "DASH"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x10  # int(0x10) = 16
    PUBKEY_ADDRESS = 0x4C  # int(0x4C) = 76  # Used to create payment addresses
    SECRET_KEY = 0xCC      # int(0xCC) = 204  # Used for WIF format
    EXT_PUBLIC_KEY = 0X0488B21E  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0X0488ADE4  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/5'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class DashTestNet(CryptoNetwork):
    """Dash TestNet version bytes."""
    NAME = "Dash"
    INTERNAL_NAME = wallet_pb2.Network.DASH_TESTNET
    COIN = "DASH"
    TESTNET = True
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x13  # int(0x13) = 19
    PUBKEY_ADDRESS = 0x8C  # int(0x8C) = 140  # Used to create payment addresses
    SECRET_KEY = 0xEF      # int(0xEF) = 239  # Used for WIF format
    EXT_PUBLIC_KEY = 0x3a805837  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x3a8061a0  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/1'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported

class DashInvertedTestNet(CryptoNetwork):
    """Dash TestNet version bytes with inverted extended version bytes."""
    NAME = "Dash"
    INTERNAL_NAME = wallet_pb2.Network.DASH_INVERTED_TESTNET
    COIN = "DASH"
    TESTNET = True
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x13  # int(0x13) = 19
    PUBKEY_ADDRESS = 0x8C  # int(0x8C) = 140  # Used to create payment addresses
    SECRET_KEY = 0xEF      # int(0xEF) = 239  # Used for WIF format
    EXT_PUBLIC_KEY = 0x3a8061a0  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x3a805837  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/1'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class BitcoinMainNet(CryptoNetwork):
    """Bitcoin MainNet version bytes defaulting to legacy addresses.
    From https://github.com/bitcoin/bitcoin/blob/v0.9.0rc1/src/chainparams.cpp
    """
    NAME = "Bitcoin"
    INTERNAL_NAME = wallet_pb2.Network.BITCOIN_MAINNET
    COIN = "BTC"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x05  # int(0x05) = 5
    PUBKEY_ADDRESS = 0x00  # int(0x00) = 0  # Used to create payment addresses
    SECRET_KEY = 0x80      # int(0x80) = 128  # Used for WIF format

    EXT_PUBLIC_KEY = 0x0488B21E  # Used to serialize public keys in BIP32 legacy (P2PKH)
    EXT_SECRET_KEY = 0x0488ADE4  # Used to serialize private keys in BIP32 legacy (P2PKH)
    BIP32_PATH = "m/44'/0'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None
    BECH32_PREFIX = None

class BitcoinSegwitMainNet(BitcoinMainNet):
    """Bitcoin MainNet version bytes, defaulting to segwit addresses."""
    INTERNAL_NAME = wallet_pb2.Network.BITCOIN_SEGWIT_MAINNET
    ADDRESS_MODE = ["BECH32", "BASE58"]
    SUPPORTS_SEGWIT = True
    EXT_SEGWIT_PUBLIC_KEY = 0x04b24746 # Used to serialize public keys in BIP32 segwit (P2WPKH)
    EXT_SEGWIT_SECRET_KEY = 0x04b2430c # Used to serialize private keys in BIP32 segwit (P2WPKH)
    BIP32_SEGWIT_PATH = "m/84'/0'/0'"
    BECH32_PREFIX = "bc"

class BitcoinTestNet(CryptoNetwork):
    """Bitcoin TestNet version bytes, defaulting to legacy addresses.
    From https://github.com/bitcoin/bitcoin/blob/v0.9.0rc1/src/chainparams.cpp
    """
    NAME = "Bitcoin"
    INTERNAL_NAME = wallet_pb2.Network.BITCOIN_TESTNET
    COIN = "BTC"
    TESTNET = True
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0xc4  # int(0xc4) = 196
    PUBKEY_ADDRESS = 0x6f  # int(0x6f) = 111
    SECRET_KEY = 0xEF      # int(0xef) = 239

    EXT_PUBLIC_KEY = 0x043587CF # Used to serialize public keys in BIP32 legacy (P2PKH)
    EXT_SECRET_KEY = 0x04358394 # Used to serialize private keys in BIP32 legacy (P2PKH)
    BIP32_PATH = "m/44'/1'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None
    BECH32_PREFIX = None

class BitcoinSegwitTestNet(BitcoinTestNet):
    """Bitcoin TestNet version bytes, defaulting to segwit addresses."""
    INTERNAL_NAME = wallet_pb2.Network.BITCOIN_SEGWIT_TESTNET
    ADDRESS_MODE = ["BECH32", "BASE58"]
    SUPPORTS_SEGWIT = True
    EXT_SEGWIT_PUBLIC_KEY = 0x045f1cf6 # Used to serialize public keys in BIP32 segwit (P2WPKH)
    EXT_SEGWIT_SECRET_KEY = 0x045f18bc # Used to serialize private keys in BIP32 segwit (P2WPKH)
    BIP32_SEGWIT_PATH = "m/84'/1'/0'"
    BECH32_PREFIX = "tb"

class LitecoinMainNet(CryptoNetwork):
    """Litecoin MainNet version bytes

    Primary version bytes from:
    https://github.com/litecoin-project/litecoin/blob/master-0.8/src/base58.h

    Extemded version bytes from https://bitcointalk.org/index.php?topic=453395.0
    """
    NAME = "Litecoin"
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_MAINNET
    COIN = "LTC"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x05  # int(0x05) = 5
    PUBKEY_ADDRESS = 0x30  # int(0x30) = 48
    SECRET_KEY = PUBKEY_ADDRESS + 128  # = int(0xb0) = 176

    # Litecoin is using xpub/xpriv - same as Bitcoin
    # According to [1] and [2], Litecoin was supposed to use Lpub/Lprv
    # but Litecoin devs never got around to implementing that.
    # [1]: https://www.reddit.com/r/litecoin/comments/7rorqa/electrum_using_bitcoin_xpub_headers/dszq9d5/?context=3
    # [2]: https://github.com/ranaroussi/pywallet/issues/6
    EXT_PUBLIC_KEY = 0x019da462
    EXT_SECRET_KEY = 0x019d9cfe
    BIP32_PATH = "m/44'/2'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None
    BECH32_PREFIX = None


class LitecoinSegwitMainNet(LitecoinMainNet):
    """Litecoin MainNet version bytes, defaulting to segwit addresses."""
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_SEGWIT_MAINNET
    ADDRESS_MODE = ["BECH32", "BASE58"]
    SUPPORTS_SEGWIT = True
    EXT_SEGWIT_PUBLIC_KEY = 0x04b24746 # Used to serialize public keys in BIP32 segwit (P2WPKH)
    EXT_SEGWIT_SECRET_KEY = 0x04b2430c # Used to serialize private keys in BIP32 segwit (P2WPKH)
    BIP32_SEGWIT_PATH = "m/84'/0'/0'"
    BECH32_PREFIX = "ltc"

class LitecoinBTCMainNet(CryptoNetwork):
    """Litecoin MainNet version bytes

    Primary version bytes from:
    https://github.com/litecoin-project/litecoin/blob/master-0.8/src/base58.h

    Extended version bytes same as bitcoin's
    """
    NAME = "Litecoin"
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_BTC_MAINNET
    COIN = "LTC"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x05  # int(0x05) = 5
    PUBKEY_ADDRESS = 0x30  # int(0x30) = 48
    SECRET_KEY = PUBKEY_ADDRESS + 128  # = int(0xb0) = 176

    # Litecoin is using xpub/xpriv - same as Bitcoin
    # According to [1] and [2], Litecoin was supposed to use Lpub/Lprv
    # but Litecoin devs never got around to implementing that.
    # Although some wallets such as Trezor are still implementing Ltub/Ltpv
    # [1]: https://www.reddit.com/r/litecoin/comments/7rorqa/electrum_using_bitcoin_xpub_headers/dszq9d5/?context=3
    # [2]: https://github.com/ranaroussi/pywallet/issues/6
    EXT_PUBLIC_KEY = 0x0488B21E
    EXT_SECRET_KEY = 0x0488ADE4
    BIP32_PATH = "m/44'/2'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None
    BECH32_PREFIX = None

class LitecoinBTCSegwitMainNet(LitecoinBTCMainNet):
    """Litecoin MainNet version bytes, defaulting to segwit addresses."""
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_BTC_SEGWIT_MAINNET
    ADDRESS_MODE = ["BECH32", "BASE58"]
    SUPPORTS_SEGWIT = True
    EXT_SEGWIT_PUBLIC_KEY = 0x04b24746 # Used to serialize public keys in BIP32 segwit (P2WPKH)
    EXT_SEGWIT_SECRET_KEY = 0x04b2430c # Used to serialize private keys in BIP32 segwit (P2WPKH)
    BIP32_SEGWIT_PATH = "m/84'/0'/0'"
    BECH32_PREFIX = "ltc"

class LitecoinTestNet(CryptoNetwork):
    """Litecoin TestNet version bytes

    Primary version bytes from:
    https://github.com/litecoin-project/litecoin/blob/master-0.8/src/base58.h

    Litecoin testnet extended version bytes for legacy are the same as for Bitcoin testnet
    so we will just reuse the segwit for Bitcoin testnet as well. There is no official document
    I could find that supports this though.
    """
    NAME = "Litecoin"
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_MAINNET
    COIN = "LTC"
    TESTNET = True
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0xc4  # int(0xc4) = 196
    PUBKEY_ADDRESS = 0x6f  # int(0x6f) = 111
    SECRET_KEY = PUBKEY_ADDRESS + 128  # = int(0xef) = 239

    EXT_PUBLIC_KEY = 0x0436f6e1
    EXT_SECRET_KEY = 0x0436ef7d
    BIP32_PATH = "m/44'/1'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None
    BECH32_PREFIX = None

class LitecoinSegwitTestNet(LitecoinTestNet):
    """Litecoin TestNet version bytes, defaulting to segwit addresses."""
    INTERNAL_NAME = wallet_pb2.Network.LITECOIN_MAINNET
    ADDRESS_MODE = ["BECH32", "BASE58"]
    SUPPORTS_SEGWIT = True
    EXT_SEGWIT_PUBLIC_KEY = 0x045f1cf6 # Used to serialize public keys in BIP32 segwit (P2WPKH)
    EXT_SEGWIT_SECRET_KEY = 0x045f18bc # Used to serialize private keys in BIP32 segwit (P2WPKH)
    BIP32_SEGWIT_PATH = "m/84'/1'/0'"
    BECH32_PREFIX = "tltc"

class DogecoinMainNet(CryptoNetwork):
    """Dogecoin MainNet version bytes

    Primary version bytes from:
    https://github.com/dogecoin/dogecoin/blob/1.5.2/src/base58.h

    Unofficial extended version bytes from
    https://bitcointalk.org/index.php?topic=409731
    """
    NAME = "Dogecoin"
    INTERNAL_NAME = wallet_pb2.Network.DOGECOIN_MAINNET
    COIN = "DOGE"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x16  # int(0x16) = 22
    PUBKEY_ADDRESS = 0x1e  # int(0x1e) = 30
    SECRET_KEY = PUBKEY_ADDRESS + 128  # int(0x9e) = 158

    # Unofficial extended version bytes taken from
    # https://bitcointalk.org/index.php?topic=409731
    # and https://github.com/dogecoin/dogecoin/blob/3a29ba6d497cd1d0a32ecb039da0d35ea43c9c85/src/chainparams.cpp
    EXT_PUBLIC_KEY = 0x02facafd
    EXT_SECRET_KEY = 0x02fac398
    BIP32_PATH = "m/44'/3'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class DogecoinBTCMainNet(CryptoNetwork):
    """Dogecoin MainNet version bytes

    Primary version bytes from:
    https://github.com/dogecoin/dogecoin/blob/1.5.2/src/base58.h

    Extended version bytes are the same as for Bitocin mainnet,
    i.e. xpub/xprv, for wallet compatibility.
    """
    NAME = "Dogecoin"
    INTERNAL_NAME = wallet_pb2.Network.DOGECOIN_BTC_MAINNET
    COIN = "DOGE"
    TESTNET = False
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x16  # int(0x16) = 22
    PUBKEY_ADDRESS = 0x1e  # int(0x1e) = 30
    SECRET_KEY = PUBKEY_ADDRESS + 128  # int(0x9e) = 158

    # Unofficial extended version bytes taken from
    # https://bitcointalk.org/index.php?topic=409731
    # and https://github.com/dogecoin/dogecoin/blob/3a29ba6d497cd1d0a32ecb039da0d35ea43c9c85/src/chainparams.cpp
    EXT_PUBLIC_KEY = 0x02facafd
    EXT_SECRET_KEY = 0x02fac398
    BIP32_PATH = "m/44'/3'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class DogecoinTestNet(CryptoNetwork):
    """Dogecoin TestNet version bytes

    Primary version bytes from:
    https://github.com/dogecoin/dogecoin/blob/1.5.2/src/base58.h

    Unofficial extended version bytes from
    https://bitcointalk.org/index.php?topic=409731
    """
    NAME = "Dogecoin"
    INTERNAL_NAME = wallet_pb2.Network.DOGECOIN_TESTNET
    COIN = "DOGE"
    TESTNET = True
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0xc4  # int(0xc4) = 196
    PUBKEY_ADDRESS = 0x71  # int(0x71) = 113
    SECRET_KEY = PUBKEY_ADDRESS + 128  # int(0xf1) = 241

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    # Unofficial extended version bytes taken from
    # https://bitcointalk.org/index.php?topic=409731
    EXT_PUBLIC_KEY = 0x0432a9a8
    EXT_SECRET_KEY = 0x0432a243
    BIP32_PATH = "m/44'/1'/0'"

    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported


class BlockcypherTestNet(CryptoNetwork):
    """Blockcypher TestNet version bytes.
    From http://dev.blockcypher.com/#testing
    """
    NAME = "BlockCypher"
    INTERNAL_NAME = wallet_pb2.Network.BLOCKCYPHER_TESTNET
    COIN = "BCY"
    ADDRESS_MODE = ["BASE58"]
    SCRIPT_ADDRESS = 0x1f  # int(0x1f) = 31
    PUBKEY_ADDRESS = 0x1b  # int(0x1b) = 27  # Used to create payment addresses
    SECRET_KEY = 0x49      # int(0x49) = 73  # Used for WIF format
    EXT_PUBLIC_KEY = 0x2d413ff  # Used to serialize public BIP32 addresses
    EXT_SECRET_KEY = 0x2d40fc3  # Used to serialize private BIP32 addresses
    BIP32_PATH = "m/44'/1'/0'"

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = False
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported

class EthereumMainNet(CryptoNetwork):
    """Ethereum MainNet version bytes."""
    NAME = "Ethereum"
    INTERNAL_NAME = wallet_pb2.Network.ETHEREUM_MAINNET
    COIN = "ETH"
    TESTNET = False
    ADDRESS_MODE = ["HEX"]

    # Ethereum doesn't put version bytes in front of keys or addresses.
    SCRIPT_ADDRESS = None
    PUBKEY_ADDRESS = None
    SECRET_KEY = None

    EXT_PUBLIC_KEY = None  # Ethereum does not use BIP32
    EXT_SECRET_KEY = None  # Ethereum does not use BIP32
    BIP32_PATH = None  # Ethereum does not use BIP32

    SUPPORTS_SEGWIT = False
    SUPPORTS_EVM = True
    EXT_SEGWIT_PUBLIC_KEY = None
    EXT_SEGWIT_SECRET_KEY = None
    BIP32_SEGWIT_PATH = None # P2WPKH not supported
    BECH32_PREFIX = None # Bech32 not supported

#TODO support Monero, ZCash



# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
eth_nodes = [
    {
        'url': 'https://1rpc.io/eth'
    },
    {
        'url': 'https://ethereum-mainnet-rpc.allthatnode.com'
    },
    {
        'url': 'https://ethereum.publicnode.com'
    },
    {
        'url': 'https://rpc.ankr.com/eth'
    },
    {
        'url': 'https://eth-mainnet.public.blastapi.io'
    },
    {
        'url': 'https://ethereum.blinklabs.xyz/'
    },
    {
        'url': 'https://api.blockeden.xyz/eth/8UuXzatAZYDBJC6YZTKD'
    },
    {
        'url': 'https://ethereum.blockpi.network/v1/rpc/public'
    },
    {
        'url': 'https://eth-protect.rpc.blxrbdn.com'
    },
    {
        'url': 'https://cloudflare-eth.com/v1/mainnet'
    },
    {
        'url': 'https://eth.drpc.org/'
    },
    {
        'url': 'https://api.edennetwork.io/v1/rocket'
    },
    {
        'url': 'https://ethereum-bundler.etherspot.io/'
    },
    {
        'url': 'https://ethereum-api.flare.network/'
    },
    {
        'url': 'https://rpc.flashbots.net/'
    },
    {
        'url': 'https://rpc.eth.gateway.fm'
    },
    {
        'url': 'https://eth.meowrpc.com'
    },
    {
        'url': 'https://eth.llamarpc.com'
    },
    {
        'url': 'https://eth.merkle.io/'
    },
    {
        'url': 'https://rpc.mevblocker.io'
    },
    {
        'url': 'https://www.noderpc.xyz/rpc-mainnet/public'
    },
    {
        'url': 'https://lb.nodies.app/v1/5e9daed367d1454fab7c75f0ec8aceff'
    },
    {
        'url': 'https://endpoints.omniatech.io/v1/eth/mainnet/public'
    },
    {
        'url': 'https://eth.api.onfinality.io/public'
    },
    {
        'url': 'https://eth-pokt.nodies.app/'
    },
    {
        'url': 'https://rpc.propellerheads.xyz/eth'
    },
    {
        'url': 'https://mainnet.gateway.tenderly.co'
    },
    {
        'url': 'https://ethereum.rpc.thirdweb.com/'
    },
]



from web3 import Web3

from ...utils.utils import eth_transaction_hash
from ...generated import wallet_pb2
from ...utils.keccak import to_checksum_address

class EthereumWeb3Client:
    """ Address querying class for Ethereum full nodes and Web3 clients.
        Most 3rd party providers e.g. Infura, QuickNode will also work here.
 
        WARNING: Ethereum nodes have a --txlookuplimit and maintain the last N transactions only,
        unless this option is turned off. 3rd party providers should have this switched off, but
        ensure it is turned off if you are running your own node.
    """
    
    def _clean_tx(self, element, block):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['hash']
        if 'blockNumber' in element.keys():
            new_element.confirmed = True
            new_element.height = element['blockNumber']
        else:
            new_element.confirmed = False

        new_element.ethlike_transaction.txfrom = element['from']
        new_element.ethlike_transaction.txto = element['to']
        new_element.ethlike_transaction.amount = int(element['value'])
        
        new_element.timestamp = hex(block["timestamp"], 16)
        new_element.ethlike_transaction.data = element['input']

        gas = int(element['gas'], 16)
        new_element.ethlike_transaction.gas = gas
        if 'maxFeePerGas' in element.keys():
            new_element.total_fee = int(element['maxFeePerGas'], 16) * gas
        else:
            new_element.total_fee = int(element['gasPrice']) * gas

        new_element.fee_metric = wallet_pb2.WEI
        return new_element

    def __init__(self, addresses, transactions=None, **kwargs):
        self.web3 = Web3(Web3.HTTPProvider(kwargs.get('url')))

        self.transactions = []
        self.addresses = [to_checksum_address(a) for a in addresses]
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.height = self.get_block_height()
        self.transactions = [*self._get_transaction_history()]

    def get_block_height(self):
        return self.web3.eth.block_number
    
    def get_balance(self):
        """
        Retrieves the balance of the Ethereum address.

        The ETH balance can be obtained without fetching the Ethereum transactions first.

        Returns:
            int: The balance of the Ethereum address in Gwei.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        balance = 0
        for address in self.addresses:
            balance += self.web3.eth.get_balance(address)
        
        # Ethereum has no unconfirmed balances or transactions.
        # But for compatibility reasons, we still return it as a 2-tuple.
        return (balance, balance)
                
    # In Ethereum, only one transaction per account can be included in a block at a time.
    def _get_transaction_history(self, block_height=None):

        addresses = [a.lower() for a in self.addresses]

        for block_number in range(block_height, self.get_block_height() + 1):
            # Retrieve block information
            block = self.web3.eth.getBlock(block_number, full_transactions=True)

            # Check if the block contains transactions
            if block and 'transactions' in block:
                transactions = block['transactions']

                # Iterate through transactions in the block
                for tx_hash in transactions:
                    # Retrieve transaction details
                    transaction = self.web3.eth.getTransaction(tx_hash)

                    # Check if the transaction is related to the target address
                    if transaction and transaction['to'].lower() in addresses:
                        yield self._clean_tx(transaction, block)


#!/usr/bin/env python3
# Copyright (c) 2019 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Utility functions related to output descriptors"""

import re

INPUT_CHARSET = "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"
GENERATOR = [0xf5dee51989, 0xa9fdca3312, 0x1bab10e32d, 0x3706b1677a, 0x644d626ffd]

def descsum_polymod(symbols):
    """Internal function that computes the descriptor checksum."""
    chk = 1
    for value in symbols:
        top = chk >> 35
        chk = (chk & 0x7ffffffff) << 5 ^ value
        for i in range(5):
            chk ^= GENERATOR[i] if ((top >> i) & 1) else 0
    return chk

def descsum_expand(s):
    """Internal function that does the character to symbol expansion"""
    groups = []
    symbols = []
    for c in s:
        if not c in INPUT_CHARSET:
            return None
        v = INPUT_CHARSET.find(c)
        symbols.append(v & 31)
        groups.append(v >> 5)
        if len(groups) == 3:
            symbols.append(groups[0] * 9 + groups[1] * 3 + groups[2])
            groups = []
    if len(groups) == 1:
        symbols.append(groups[0])
    elif len(groups) == 2:
        symbols.append(groups[0] * 3 + groups[1])
    return symbols

def descsum_create(s):
    """Add a checksum to a descriptor without"""
    symbols = descsum_expand(s) + [0, 0, 0, 0, 0, 0, 0, 0]
    checksum = descsum_polymod(symbols) ^ 1
    return s + '#' + ''.join(CHECKSUM_CHARSET[(checksum >> (5 * (7 - i))) & 31] for i in range(8))

def descsum_create_only(s):
    """Returns only the checksum of the descriptor"""
    symbols = descsum_expand(s) + [0, 0, 0, 0, 0, 0, 0, 0]
    checksum = descsum_polymod(symbols) ^ 1
    return ''.join(CHECKSUM_CHARSET[(checksum >> (5 * (7 - i))) & 31] for i in range(8))

def descsum_check(s, require=True):
    """Verify that the checksum is correct in a descriptor"""
    if not '#' in s:
        return not require
    if s[-9] != '#':
        return False
    if not all(x in CHECKSUM_CHARSET for x in s[-8:]):
        return False
    symbols = descsum_expand(s[:-9]) + [CHECKSUM_CHARSET.find(x) for x in s[-8:]]
    return descsum_polymod(symbols) == 1

def drop_origins(s):
    '''Drop the key origins from a descriptor'''
    desc = re.sub(r'\[.+?\]', '', s)
    if '#' in s:
        desc = desc[:desc.index('#')]
    return descsum_create(desc)


# -*- coding: utf-8 -*-
"""
This module contains the methods for creating a crypto wallet.
"""

from os import urandom
from random import randrange
from typing import List

from zpywallet.utxo import UTXO

from .destination import Destination

from .mnemonic import Mnemonic
from .utils.bip32 import (
    HDWallet
)

from .utils.keys import (
    PrivateKey,
    PublicKey
)
from .transactions.encode import create_transaction
from .transactions.decode import transaction_size_simple
from .broadcast import broadcast_transaction

from .generated import wallet_pb2

from .network import (
    BitcoinSegwitMainNet,
    BitcoinMainNet,
    BitcoinSegwitTestNet,
    BitcoinTestNet,
    LitecoinSegwitMainNet,
    LitecoinMainNet,
    LitecoinBTCSegwitMainNet,
    LitecoinBTCMainNet,
    LitecoinSegwitTestNet,
    LitecoinTestNet,
    EthereumMainNet,
    DogecoinMainNet,
    DogecoinBTCMainNet,
    DogecoinTestNet,
    DashMainNet,
    DashInvertedTestNet,
    DashBTCMainNet,
    DashTestNet,
    DashInvertedMainNet,
    BitcoinCashMainNet,
    BlockcypherTestNet
)

from .address.bcy import BCYAddress
from .address.btc import BitcoinAddress
from .address.btctest import BitcoinTestAddress
from .address.dash import DashAddress
from .address.doge import DogecoinAddress
from .address.eth import EthereumAddress
from .address.ltc import LitecoinAddress

from .nodes.eth import eth_nodes

from .utils.aes import decrypt, encrypt

from .transaction import Transaction

def generate_mnemonic(strength=128):
    """Creates a new seed phrase of the specified length"""
    if strength % 32 != 0:
        raise ValueError("strength must be a multiple of 32")
    if strength < 128 or strength > 256:
        raise ValueError("strength should be >= 128 and <= 256")
    entropy = urandom(strength // 8)
    mne = Mnemonic(language='english')
    mnemonic = mne.to_mnemonic(entropy)
    return mnemonic


def create_wallet(mnemonic=None, network=BitcoinSegwitMainNet, strength=128) -> HDWallet:
    """Generate a new wallet class from a mnemonic phrase, optionally randomly generated

    Args:
    :param mnemonic: The key to use to generate this wallet. It may be a long
        string. Do not use a phrase from a book or song, as that will
        be guessed and is not secure. My advice is to not supply this
        argument and let me generate a new random key for you.
    :param network: The network to create this wallet for
    :param children: Create this many child addresses for this wallet. Default
        is 10, You should not use the master private key itself for sending
        or receiving funds, as a best practice.

    Return:
        HDWallet: a wallet class
    
    Usage:
        w = create_wallet(network='BTC', children=10)
    """
    if mnemonic is None:
        return HDWallet.from_random(strength=strength, network=network)
    else:
        return HDWallet.from_mnemonic(mnemonic=mnemonic, network=network)


def create_keypair(network=BitcoinSegwitMainNet):
    """Generates a random private/public keypair.

    Args:
    :param network: The network to create this wallet for

    Return:
        PrivateKey, PublicKey: a tuple of a private key and public key.
    
    Usage:
        w = create_wallet(network='BTC', children=10)
    """

    random_bytes = urandom(32)
    prv = PrivateKey(random_bytes, network=network)
    pub = prv.public_key
    return prv, pub

class Wallet:
    """Data class representing a cryptocurrency wallet."""

    def __init__(self, network, seed_phrase, password, receive_gap_limit=1000, change_gap_limit=1000,
                  derivation_path=None, _with_wallet=True, **kwargs):
        
        fullnode_endpoints = kwargs.get('fullnode_endpoints')
        esplora_endpoints = kwargs.get('esplora_endpoints')
        blockcypher_tokens = kwargs.get('blockcypher_tokens')

        self._network = network
        if not derivation_path:
            derivation_path = network.BIP32_SEGWIT_PATH if network.BIP32_SEGWIT_PATH else network.BIP32_PATH

        if not seed_phrase:
            seed_phrase = generate_mnemonic()
        
        if _with_wallet:
            self.wallet = wallet_pb2.Wallet()
            self.wallet.SerializeToString()
            self.wallet.receive_gap_limit = receive_gap_limit
            self.wallet.change_gap_limit = change_gap_limit
            self.wallet.height = 0

            if type(derivation_path) is str:
                self.wallet.derivation_path = derivation_path
            else:
                raise ValueError("Invalid derivation path")
            
            # Generate addresses and keys
            hdwallet = HDWallet.from_mnemonic(mnemonic=seed_phrase, network=network)
            
            # We do not save the password. Instead, we are going to generate a base64-encrypted
            # serialization of this wallet file using the password.
            self.wallet.encrypted_seed_phrase = encrypt(seed_phrase, password) # AES-256-CBC encryption

            # Set properties
            if network == BitcoinSegwitMainNet:
                self.wallet.network = wallet_pb2.BITCOIN_SEGWIT_MAINNET
            elif network == BitcoinMainNet:
                self.wallet.network = wallet_pb2.BITCOIN_MAINNET
            elif network == BitcoinSegwitTestNet:
                self.wallet.network = wallet_pb2.BITCOIN_SEGWIT_TESTNET
            elif network == BitcoinTestNet:
                self.wallet.network = wallet_pb2.BITCOIN_TESTNET
            elif network == LitecoinSegwitMainNet:
                self.wallet.network = wallet_pb2.LITECOIN_SEGWIT_MAINNET
            elif network == LitecoinMainNet:
                self.wallet.network = wallet_pb2.LITECOIN_MAINNET
            elif network == LitecoinBTCSegwitMainNet:
                self.wallet.network = wallet_pb2.LITECOIN_BTC_SEGWIT_MAINNET
            elif network == LitecoinBTCMainNet:
                self.wallet.network = wallet_pb2.LITECOIN_BTC_MAINNET
            elif network == LitecoinSegwitTestNet:
                self.wallet.network = wallet_pb2.LITECOIN_SEGWIT_TESTNET
            elif network == LitecoinTestNet:
                self.wallet.network = wallet_pb2.LITECOIN_TESTNET
            elif network == EthereumMainNet:
                self.wallet.network = wallet_pb2.ETHEREUM_MAINNET
            elif network == DogecoinMainNet:
                self.wallet.network = wallet_pb2.DOGECOIN_MAINNET
            elif network == DogecoinBTCMainNet:
                self.wallet.network = wallet_pb2.DOGECOIN_BTC_MAINNET
            elif network == DogecoinTestNet:
                self.wallet.network = wallet_pb2.DOGECOIN_TESTNET
            elif network == DashMainNet:
                self.wallet.network = wallet_pb2.DASH_MAINNET
            elif network == DashInvertedMainNet:
                self.wallet.network = wallet_pb2.DASH_INVERTED_MAINNET
            elif network == DashBTCMainNet:
                self.wallet.network = wallet_pb2.DASH_BTC_MAINNET
            elif network == DashTestNet:
                self.wallet.network = wallet_pb2.DASH_TESTNET
            elif network == DashInvertedTestNet:
                self.wallet.network = wallet_pb2.DASH_INVERTED_TESTNET
            elif network == BitcoinCashMainNet:
                self.wallet.network = wallet_pb2.BITCOIN_CASH_MAINNET
            elif network == BlockcypherTestNet:
                self.wallet.network = wallet_pb2.BLOCKCYPHER_TESTNET
            else:
                raise ValueError("Unkown network")

            if fullnode_endpoints is not None:
                self.wallet.fullnode_endpoints.extend(fullnode_endpoints)

            if esplora_endpoints is not None:
                self.wallet.esplora_endpoints.extend(esplora_endpoints)

            if blockcypher_tokens is not None:
                self.wallet.blockcypher_tokens.extend(blockcypher_tokens)
            
            self.encrypted_private_keys = []
            for i in range(0, receive_gap_limit):
                privkey = hdwallet.get_child_for_path(f"{derivation_path}/0/{i}").private_key
                self.encrypted_private_keys.append(encrypt(privkey.to_hex() if network.SUPPORTS_EVM else privkey.to_wif(), password))
                pubkey = privkey.public_key

                # Add an Address
                address = self.wallet.addresses.add()
                address.address = pubkey.address()
                address.pubkey = pubkey.to_hex()
                address.privkey = privkey.to_wif()

            # for i in range(0, change_gap_limit):
            #     privkey = hdwallet.get_child_for_path(f"{derivation_path}/1/{i}").private_key
            #     pubkey = privkey.public_key
                
            #     # Add an Address
            #     address = self.wallet.addresses.add()
            #     address.address = pubkey.address()
            #     address.pubkey = pubkey.to_hex()
            #     address.privkey = privkey.to_wif()
            

    @classmethod
    def deserialize(cls, data: bytes, password):
        wallet = wallet_pb2.Wallet()
        wallet.ParseFromString(data)
        seed_phrase = decrypt(wallet.encrypted_seed_phrase, password)

        if wallet.network == wallet_pb2.BITCOIN_SEGWIT_MAINNET:
            network = BitcoinSegwitMainNet
        elif wallet.network == wallet_pb2.BITCOIN_MAINNET:
            network = BitcoinMainNet
        elif wallet.network == wallet_pb2.BITCOIN_SEGWIT_TESTNET:
            network = BitcoinSegwitTestNet
        elif wallet.network == wallet_pb2.BITCOIN_TESTNET:
            network = BitcoinTestNet
        elif wallet.network == wallet_pb2.LITECOIN_SEGWIT_MAINNET:
            network = LitecoinSegwitMainNet
        elif wallet.network == wallet_pb2.LITECOIN_MAINNET:
            network = LitecoinMainNet
        elif wallet.network == wallet_pb2.LITECOIN_BTC_SEGWIT_MAINNET:
            network = LitecoinBTCSegwitMainNet
        elif wallet.network == wallet_pb2.LITECOIN_BTC_MAINNET:
            network = LitecoinBTCMainNet
        elif wallet.network == wallet_pb2.LITECOIN_SEGWIT_TESTNET:
            network = LitecoinSegwitTestNet
        elif wallet.network == wallet_pb2.LITECOIN_TESTNET:
            network = LitecoinTestNet
        elif wallet.network == wallet_pb2.ETHEREUM_MAINNET:
            network = EthereumMainNet
        elif wallet.network == wallet_pb2.DOGECOIN_MAINNET:
            network = DogecoinMainNet
        elif wallet.network == wallet_pb2.DOGECOIN_BTC_MAINNET:
            network = DogecoinBTCMainNet
        elif wallet.network == wallet_pb2.DOGECOIN_TESTNET:
            network = DogecoinTestNet
        elif wallet.network == wallet_pb2.DASH_MAINNET:
            network = DashMainNet
        elif wallet.network == wallet_pb2.DASH_INVERTED_MAINNET:
            network = DashInvertedMainNet
        elif wallet.network == wallet_pb2.DASH_BTC_MAINNET:
            network = DashBTCMainNet
        elif wallet.network == wallet_pb2.DASH_TESTNET:
            network = DashTestNet
        elif wallet.network == wallet_pb2.DASH_INVERTED_TESTNET:
            network = DashInvertedTestNet
        elif wallet.network == wallet_pb2.BITCOIN_CASH_MAINNET:
            network = BitcoinCashMainNet
        elif wallet.network == wallet_pb2.BLOCKCYPHER_TESTNET:
            network = BlockcypherTestNet
        else:
            raise ValueError("Unkown network")

        self = cls(network, seed_phrase, password, _with_wallet=False)
        self.wallet = wallet
        hdwallet = HDWallet.from_mnemonic(mnemonic=seed_phrase, network=network)
        
        self.encrypted_private_keys = []
        for i in range(0, self.wallet.receive_gap_limit):
            privkey = hdwallet.get_child_for_path(f"{self.wallet.derivation_path}/0/{i}").private_key
            self.encrypted_private_keys.append(encrypt(privkey.to_hex() if network.SUPPORTS_EVM else privkey.to_wif(), password))
            pubkey = privkey.public_key

            # Add an Address
            address = self.wallet.addresses.add()
            address.address = pubkey.address()
            address.pubkey = pubkey.to_hex()
            address.privkey = privkey.to_wif()

        del(seed_phrase)
        del(password)

    def network(self):
        return self._network

    def get_transaction_history(self, max_cycles=100):
        addresses = [a.address for a in self.wallet.addresses]

        fullnode_endpoints = []
        esplora_endpoints = []
        blockcypher_tokens = []
        
        for node_pb2 in self.wallet.fullnode_endpoints:
            node = {}
            if node_pb2.url:
                node['url'] = node_pb2.url
            if node_pb2.user:
                node['user'] = node_pb2.user
            if node_pb2.password:
                node['password'] = node_pb2.password
            fullnode_endpoints.append(node)
        
        for node_pb2 in self.wallet.esplora_endpoints:
            node = {}
            if node_pb2.url:
                node['url'] = node_pb2.url
            esplora_endpoints.append(node)
                    
        for token in self.wallet.blockcypher_tokens:
            blockcypher_tokens.append(token)

        kwargs = {'fullnode_endpoints': fullnode_endpoints, 'esplora_endpoints': esplora_endpoints, 'blockcypher_tokens': blockcypher_tokens}

        if self._network.COIN == "BCY":
            address_client = BCYAddress(addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "BTC" and not self._network.TESTNET:
            address_client = BitcoinAddress(addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "BTC" and self._network.TESTNET:
            address_client = BitcoinTestAddress(addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "LTC" and not self._network.TESTNET:
            address_client = LitecoinAddress( addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "DOGE" and not self._network.TESTNET:
            address_client = DogecoinAddress(addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "DASH" and not self._network.TESTNET:
            address_client = DashAddress(addresses, max_cycles=max_cycles, **kwargs)
        elif self._network.COIN == "ETH" and not self._network.TESTNET:
            address_client = EthereumAddress(addresses, max_cycles=max_cycles, **kwargs)
        else:
            raise ValueError("No address client for this network")

        address_client.sync()
        transactions = address_client.get_transaction_history()
        # Create a set to keep track of unique txid values
        seen_txids = set()

        # List to store deduplicated transactions
        deduplicated_transactions = []

        # Iterate through transactions
        for transaction in transactions:
            txid = transaction.txid

            # Check if txid is not seen before
            if txid not in seen_txids:
                # Add the transaction to the deduplicated list
                deduplicated_transactions.append(transaction)

                # Mark this txid as seen
                seen_txids.add(txid)

        # Update the transactions list with deduplicated transactions
        transactions = deduplicated_transactions
        del self.wallet.transactions[:]
        self.wallet.transactions.extend(transactions)
        tx_array = []
        for t in transactions:
            tx_array.append(Transaction(t, self._network))
        return tx_array
    
    def get_utxos(self, max_cycles=100, only_unspent=False):
        addresses = [a.address for a in self.wallet.addresses]

        transactions = self.get_transaction_history(max_cycles=max_cycles)
        utxo_set = []
        for t in transactions:
            for i in range(len(t.sat_outputs(only_unspent=only_unspent))):
                try:
                    utxo_set.append(UTXO(t, i, addresses=addresses, other_transactions=transactions, only_mine=True))
                except ValueError:
                    pass
        
        return utxo_set
    
    def _to_human_friendly_utxo(self, inputs, private_keys):
        new_inputs = []
        for ii in range(len(inputs)):
            u = inputs[ii]
            for i in range (len(private_keys)):
                private_key = private_keys[i]
                privkey = PrivateKey.from_wif(private_key.decode(), self._network)
                try:
                    a = [privkey.public_key.base58_address(True),
                        privkey.public_key.base58_address(False),
                        privkey.public_key.bech32_address()]
                except Exception as e:
                    a = [privkey.public_key.base58_address(True),
                        privkey.public_key.base58_address(False)]
                u._output['private_key'] = private_key if u._output['address'] in a else None
                u._output['script_pubkey'] = PublicKey.script(u._output['address'], self._network)
                del(privkey)
                del(private_key)
                if u._output['private_key'] == None:
                    continue
                new_inputs.append(u)
                break
        return new_inputs
    
    def get_balance(self, in_standard_units=True, max_cycles=100):
        addresses = [a.address for a in self.wallet.addresses]
        
        if self._network.SUPPORTS_EVM:
            # We must use the Web3 network to get the balance as UTXOs are not available and getting transaction history
            # of an address is impractically slow.
            fullnode_endpoints = []
            
            for node_pb2 in self.wallet.fullnode_endpoints:
                node = {}
                if node_pb2.url:
                    node['url'] = node_pb2.url
                if node_pb2.user:
                    node['user'] = node_pb2.user
                if node_pb2.password:
                    node['password'] = node_pb2.password
                fullnode_endpoints.append(node)
            kwargs = {'fullnode_endpoints': fullnode_endpoints}
            address_client = EthereumAddress(addresses, transactions=self.wallet.transactions, max_cycles=max_cycles, **kwargs)
            balance = address_client.get_balance()
            if in_standard_units:
                return balance[0] / 1e18, balance[1] / 1e18
            else:
                return balance
        
        # Not an EVM chain

        total_balance = 0
        confirmed_balance = 0

        utxos = self.get_utxos(only_unspent=True)
        for u in utxos:
            confirmed_balance += u.amount(in_standard_units=in_standard_units)

        utxos = self.get_utxos()
        for u in utxos:
            total_balance += u.amount(in_standard_units=in_standard_units)

        return total_balance, confirmed_balance

    def addresses(self):
        return [a.address for a in self.wallet.addresses]
    
    def random_address(self):
        addresses = self.addresses()
        return addresses[randrange(0, len(addresses))]

    def private_keys(self, password):
        private_keys = []
        for p in self.encrypted_private_keys:
            try:
                private_keys.append(decrypt(p, password))
            except ValueError as e:
                del(private_keys)
                raise e
        return private_keys


    def _add_stock_nodes(self):
        fullnode_endpoints = []
        for f in self.wallet.fullnode_endpoints:
            _f = {}
            _f['url'] = f.url
            fullnode_endpoints.append(_f)
        if not self._network.SUPPORTS_EVM:
            return fullnode_endpoints
        else:
            if self._network.COIN == "ETH":
                fullnode_endpoints.extend(eth_nodes)
            return fullnode_endpoints
        
    def _calculate_change(self, inputs, destinations, fee_rate):
        temp_transaction = create_transaction(inputs, destinations, network=self._network)
        size = transaction_size_simple(temp_transaction)
        total_inputs = sum([i.amount(in_standard_units=False) for i in inputs])
        total_outputs = sum([o.amount(in_standard_units=False) for o in destinations])
        if total_inputs < total_outputs + size*fee_rate:
            raise ValueError("Not enough balance for this transaction")
        change = total_inputs - total_outputs - size*fee_rate
        return None if change <= 0 else Destination(self.random_address(), change / 1e8, self._network)


    # Fee rate is in the unit used by the network, ie. vbytes, bytes or wei
    def create_transaction(self, password: bytes, destinations: List[Destination], fee_rate, spend_unconfirmed_inputs=False, **kwargs):
        inputs = self.get_utxos(only_unspent=True)

        if not spend_unconfirmed_inputs:
            confirmed_inputs = []
            for i in inputs:
                if i.height():
                    confirmed_inputs.append(i)
            inputs = confirmed_inputs
        
        private_keys = self.private_keys(password)

        inputs = self._to_human_friendly_utxo(inputs, private_keys)

        fullnode_endpoints = self._add_stock_nodes()

        if self._network.SUPPORTS_EVM:
            kwargs['gas'] = fee_rate
            create_transaction(inputs, destinations, network=self._network, full_nodes=fullnode_endpoints, **kwargs)
        
        # Depending on the size of the transactions, we may need to add a change output. Otherwise,
        # the remaining balance is going to the miner.
        # This is not the real change input, we need to find the size of the transaction first.
        change = Destination(self.random_address(), 0, self._network)
        destinations_without_change = destinations[:]
        destinations_without_change.append(change)

        change = self._calculate_change(inputs, destinations_without_change, fee_rate)
        if change:
            destinations.append(change)
        return create_transaction(inputs, destinations, network=self._network)


    def broadcast_transaction(self, transaction):
        broadcast_transaction(transaction, self._network)

    def serialize(self):
        return self.wallet.SerializeToString()


from ..network import BitcoinSegwitMainNet
from .script import Script

class InvalidTransactionError(Exception):
    pass

def hex_to_int(string):
    return int.from_bytes(bytes.fromhex(string), byteorder="little")


def parse_transaction(raw_transaction_hex, segwit=False):
    transaction = {}
    witness_start = 0
    witness_end = 0
    witness_flag_size = 0

    try:
        # Version
        transaction['version'] = hex_to_int(raw_transaction_hex[0:8])
        index = 8

        # Marker & Flag (for SegWit)
        if segwit:
            marker = raw_transaction_hex[index:index+2]
            index += 2
            flag = raw_transaction_hex[index:index+2]
            index += 2
            witness_flag_size = 2

            if hex_to_int(marker) == 0 and hex_to_int(flag) != 1:
                raise InvalidTransactionError("Marker byte must be 0x00, flag byte immediately after it must be 0x01")

        # Input Count
        input_count, varint_length = parse_varint_hex(raw_transaction_hex[index:])
        transaction['input_count'] = input_count
        index += varint_length * 2

        if input_count == 0:
            raise InvalidTransactionError("Input count must not be zero (is this a segwit transaction?)")

        # Inputs
        transaction['inputs'] = []
        for _ in range(input_count):
            input_data = {}
            # Previous Transaction Hash
            b = bytearray.fromhex(raw_transaction_hex[index:index+64])
            b.reverse()
            input_data['prev_tx_hash'] = b.hex()
            index += 64

            # Previous Transaction Output Index
            input_data['prev_tx_output_index'] = hex_to_int(raw_transaction_hex[index:index+8])
            index += 8

            # Script Length
            script_length, varint_length = parse_varint_hex(raw_transaction_hex[index:])
            index += varint_length * 2

            # Script Signature
            input_data['script_signature'] = raw_transaction_hex[index:index+(script_length*2)]
            index += script_length * 2

            # Sequence
            input_data['sequence'] = raw_transaction_hex[index:index+8]
            index += 8

            transaction['inputs'].append(input_data)

        # Output Count
        output_count, varint_length = parse_varint_hex(raw_transaction_hex[index:])
        transaction['output_count'] = output_count
        index += varint_length * 2

        if output_count == 0:
            raise InvalidTransactionError("Output count must not be zero "
                                          "(If this is a segwit transaction, pass segwit=True)")

        # Outputs
        transaction['outputs'] = []
        for _ in range(output_count):
            output_data = {}
            # Value
            output_data['value'] = hex_to_int(raw_transaction_hex[index:index+16])
            index += 16

            # Script Length
            script_length, varint_length = parse_varint_hex(raw_transaction_hex[index:])
            index += varint_length * 2

            # Script Public Key
            output_data['script_pubkey'] = raw_transaction_hex[index:index+(script_length*2)]

            
            index += script_length * 2

            transaction['outputs'].append(output_data)

        # Witness Data (for SegWit)
        # Ensure that the flag signals that witness data is present.
        if segwit and flag:
            for j in range(input_count):
                witness_start = index
                transaction["inputs"][j]['witness_data'] = []
                witness_count, varint_length = parse_varint_hex(raw_transaction_hex[index:])
                index += varint_length * 2

                for _ in range(witness_count):
                    item_length, varint_length = parse_varint_hex(raw_transaction_hex[index:])
                    index += varint_length * 2
                    item = raw_transaction_hex[index:index+(item_length*2)]
                    index += item_length * 2
                    transaction["inputs"][j]['witness_data'].append(item)

            witness_end = index
        # Lock Time
        transaction['lock_time'] = raw_transaction_hex[index:index+8]

        if index+8 != len(raw_transaction_hex):
            raise InvalidTransactionError("Junk bytes after the transaction "
                            "(If this is a segwit transaction, pass segwit=True)")

        # Don't forget thhe lengths are in hex characters
        witness_size = (witness_end - witness_start) // 2 + witness_flag_size
        
        return transaction, witness_size
    
    except IndexError as e:
        raise InvalidTransactionError("Transaction too short") from e
    except ValueError as e:
        raise InvalidTransactionError("Invalid hexadecimal") from e

def parse_varint_hex(data):
    varint_type = hex_to_int(data[0:2])
    if varint_type < 0xfd:
        return varint_type, 1
    elif varint_type == 0xfd:
        return hex_to_int(data[2:6]), 3
    elif varint_type == 0xfe:
        return hex_to_int(data[2:10]), 5
    elif varint_type == 0xff:
        return hex_to_int(data[2:18]), 9

def transaction_size(raw_transaction_hex, segwit=False):
    _, witness_size = parse_transaction(raw_transaction_hex, segwit)
    if not segwit:
        # Pre-segwit transaction is just the transaction length (in bytes).
        return len(raw_transaction_hex) // 2
    else:
        # Calculate the size in weight units first
        tx_full_size = len(raw_transaction_hex) // 2

        # Weight units are:
        # tx size without witness, times 3,
        # plus the entire transaction size
        weight_units = (tx_full_size - witness_size) * 3 + tx_full_size

        # Convert to vbytes
        return round(weight_units / 4)

def transaction_size_simple(raw_transaction_hex):
    """ Returns the transaction size without throwing exceptions for valid & modern Segwit transactions """
    try:
        return transaction_size(raw_transaction_hex, False)
    except InvalidTransactionError:
        return transaction_size(raw_transaction_hex, True)


def parse_transaction_simple(raw_transaction_hex):
    """ Returns the parsed raw transaction without throwing exceptions for valid & modern Segwit transactions """
    try:
        return parse_transaction(raw_transaction_hex, False)
    except InvalidTransactionError:
        return parse_transaction(raw_transaction_hex, True)

def insert_address_in_outputs(fine_rawtx, network=BitcoinSegwitMainNet):
    for i in range(len(fine_rawtx["outputs"])):
        out = fine_rawtx["outputs"][i]
        script = Script.from_raw(out["script_pubkey"], network=network) # I have no idea what to use the has_segwit flag for.
        fine_rawtx["outputs"][i]["address"] = script.to_p2pkh() or script.to_p2sh() or script.to_p2wpkh() \
                                            or script.to_p2wsh() or script.to_p2tr()
    return fine_rawtx




from itertools import product
from random import getrandbits
import pytest
from zpywallet.utils.base58 import (
    b58encode, b58decode, b58encode_check, b58decode_check, b58encode_int,
    b58decode_int,
    BITCOIN_ALPHABET,
    XRP_ALPHABET)


BASE45_ALPHABET = b"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:"


@pytest.fixture(params=[BITCOIN_ALPHABET, XRP_ALPHABET, BASE45_ALPHABET])
def alphabet(request) -> str:
    return request.param


def test_simple_encode():
    data = b58encode(b'hello world')
    assert data == b'StV1DL6CwTryKyV'


def test_leadingz_encode():
    data = b58encode(b'\0\0hello world')
    assert data == b'11StV1DL6CwTryKyV'


def test_encode_empty():
    data = b58encode(b'')
    assert data == b''


def test_simple_decode():
    data = b58decode('StV1DL6CwTryKyV')
    assert data == b'hello world'


def test_simple_decode_bytes():
    data = b58decode(b'StV1DL6CwTryKyV')
    assert data == b'hello world'


def test_autofix_decode_bytes():
    data = b58decode(b'StVlDL6CwTryKyV', autofix=True)
    assert data == b'hello world'


def test_leadingz_decode():
    data = b58decode('11StV1DL6CwTryKyV')
    assert data == b'\0\0hello world'


def test_leadingz_decode_bytes():
    data = b58decode(b'11StV1DL6CwTryKyV')
    assert data == b'\0\0hello world'


def test_empty_decode():
    data = b58decode('1')
    assert data == b'\0'


def test_empty_decode_bytes():
    data = b58decode(b'1')
    assert data == b'\0'


def test_check_str():
    data = 'hello world'
    out = b58encode_check(data)
    assert out == b'3vQB7B6MrGQZaxCuFg4oh'
    back = b58decode_check(out)
    assert back == b'hello world'


def test_autofix_check_str():
    data = '3vQB7B6MrGQZaxCuFg4Oh'
    back = b58decode_check(data, autofix=True)
    assert back == b'hello world'


def test_autofix_not_applicable_check_str():
    charset = BITCOIN_ALPHABET.replace(b'x', b'l')
    msg = b'hello world'
    enc = b58encode_check(msg).replace(b'x', b'l').replace(b'o', b'0')
    back = b58decode_check(enc, alphabet=charset, autofix=True)
    assert back == msg


def test_check_failure():
    data = '3vQB7B6MrGQZaxCuFg4oH'
    with pytest.raises(ValueError):
        b58decode_check(data)


def test_check_identity(alphabet):
    data = b'hello world'
    out = b58decode_check(
        b58encode_check(data, alphabet=alphabet),
        alphabet=alphabet
    )
    assert out == data


def test_round_trips(alphabet):
    possible_bytes = [b'\x00', b'\x01', b'\x10', b'\xff']
    for length in range(0, 5):
        for bytes_to_test in product(possible_bytes, repeat=length):
            bytes_in = b''.join(bytes_to_test)
            bytes_out = b58decode(
                b58encode(bytes_in, alphabet=alphabet),
                alphabet=alphabet)
            assert bytes_in == bytes_out


def test_simple_integers(alphabet):
    for idx, char in enumerate(alphabet):
        charbytes = bytes([char])
        assert b58decode_int(charbytes, alphabet=alphabet) == idx
        assert b58encode_int(idx, alphabet=alphabet) == charbytes


def test_large_integer():
    number = 0x111d38e5fc9071ffcd20b4a763cc9ae4f252bb4e48fd66a835e252ada93ff480d6dd43dc62a641155a5  # noqa
    assert b58decode_int(BITCOIN_ALPHABET) == number
    assert b58encode_int(number) == BITCOIN_ALPHABET[1:]


def test_invalid_input():
    data = 'xyz\b'   # backspace is not part of the bitcoin base58 alphabet
    with pytest.raises(ValueError, match="Invalid character '\\\\x08'"):
        b58decode(data)


@pytest.mark.parametrize('length', [8, 32, 256, 1024])
def test_encode_random(benchmark, length) -> None:
    data = getrandbits(length * 8).to_bytes(length, byteorder='big')
    encoded = benchmark(lambda: b58encode(data))
    assert b58decode(encoded) == data


@pytest.mark.parametrize('length', [8, 32, 256, 1024])
def test_decode_random(benchmark, length) -> None:
    origdata = getrandbits(length * 8).to_bytes(length, byteorder='big')
    encoded = b58encode(origdata)
    data = benchmark(lambda: b58decode(encoded))
    assert data == origdata


import requests
import time

from ...errors import NetworkException
from ...generated import wallet_pb2

class BTCDotComAddress:
    """
    A class representing a Bitcoin address.

    This class allows you to retrieve the balance and transaction history of a Bitcoin address using the BTC.com API.

    Args:
        api_key (str): The API key for accessing the BTC.com API.

    Attributes:
        api_key (str): The API key for accessing the BTC.com API.

    Methods:
        get_balance(): Retrieves the balance of the Bitcoin address.
        get_transaction_history(): Retrieves the transaction history of the Bitcoin address.

    Raises:
        Exception: If the API request fails or the address balance/transaction history cannot be retrieved.
    """

    def _clean_tx(self, element):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['hash']
        if 'block_height' in element.keys() and element['block_height'] is not None:
            new_element.confirmed = True
            new_element.height = element['block_height']
        else:
            new_element.confirmed = False
        
        if new_element.confirmed:
            new_element.timestamp = element['block_time']

        for vin in element['inputs']:
            txinput = new_element.btclike_transaction.inputs.add()
            txinput.txid = vin['prev_tx_hash']
            txinput.index = vin['prev_position']
            txinput.amount = int(vin['prev_value'])

        i = 0 
        for vout in element['outputs']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(vout['value'])
            txoutput.index = i
            i += 1
            if 'addresses' in vout.keys():
                txoutput.address = vout['addresses'][0]
            txoutput.spent = vout['spent_by_tx'] != ""

        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])
        new_element.total_fee = (total_inputs - total_outputs)

        new_element.btclike_transaction.fee = int((total_inputs - total_outputs) // element['vsize'])
        new_element.fee_metric = wallet_pb2.VBYTE
        return new_element

    # BTC.com's rate limits are unknown.
    def __init__(self, addresses, request_interval=(1000,1), transactions=None):
        """
        Initializes an instance of the BTCDotComAddress class.

        Args:
            addresses (list): A list of human-readable Bitcoin addresses.
            request_interval (tuple): A pair of integers indicating the number of requests allowed during
                a particular amount of seconds. Set to (0,N) for no rate limiting, where N>0.
        """
        self.requests, self.interval_sec = request_interval
        self.addresses = addresses
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.transactions = [*self._get_transaction_history()]
        self.height = self.get_block_height()

    def get_balance(self):
        """
        Retrieves the balance of the Bitcoin address.

        Returns:
            float: The balance of the Bitcoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos

    def get_block_height(self):
        # Get the current block height now:
        url = "https://chain.api.btc.com/v3/block/latest"
        for attempt in range(3, -1, -1):
            if attempt == 0:
                raise NetworkException("Network request failure")
            try:
                response = requests.get(url, timeout=60)
                break
            except requests.RequestException:
                pass
            except requests.exceptions.JSONDecodeError:
                pass
        if response.status_code == 200:
            data = response.json()
            self.height = data["data"]["height"]
        else:
            try:
                return self.height
            except AttributeError as exc:
                raise NetworkException("Failed to retrieve current blockchain height") from exc

        
    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Bitcoin address from cached data augmented with network data.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs
                    
        return self.transactions

    def _get_transaction_history(self, txhash=None):
        """
        Retrieves the transaction history of the Bitcoin address. (internal method that makes the network query)

        Parameters:
            txhash (str): Get all transactions before (and not including) txhash.
                Defaults to None, which disables this behavior.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        for address in self.addresses:
            page = 1
            pagesize = 50

            url = f"https://chain.api.btc.com/v3/address/{address}/tx?page={page}&pagesize={pagesize}"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, timeout=60)
                    if response.status_code == 200:
                        data = response.json()
                        break
                    else:
                        raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass

                if data["data"]["list"] is None:
                    return
                for tx in data["data"]["list"]:
                    time.sleep(self.interval_sec/(self.requests*len(data["data"]["list"])))
                    if txhash and tx["hash"] == txhash:
                        return
                    yield self._clean_tx(tx)
                page += 1
            else:
                raise NetworkException("Failed to retrieve transaction history")
            

            while data["data"]["list"] is not None: 
                url = f"https://chain.api.btc.com/v3/address/{address}/tx?page={page}&pagesize={pagesize}"
                for attempt in range(3, -1, -1):
                    if attempt == 0:
                        raise NetworkException("Network request failure")
                    try:
                        response = requests.get(url, timeout=60)
                        break
                    except requests.RequestException:
                        pass

                if response.status_code == 200:
                    data = response.json()
                    if data["data"]["list"] is None:
                        return
                    for tx in data["data"]["list"]:
                        time.sleep(self.interval_sec/(self.requests*len(data["data"]["list"])))
                        if txhash and tx["hash"] == txhash:
                            return
                        yield self._clean_tx(tx)
                    page += 1
                else:
                    raise NetworkException("Failed to retrieve transaction history")



from .generated import wallet_pb2

class Transaction:
    def __init__(self, transaction: wallet_pb2.Transaction, network):
        self._network = network
        self._txid = transaction.txid
        self._timestamp = transaction.timestamp
        self._confirmed = transaction.confirmed
        self._height = transaction.height
        self._total_fee = transaction.total_fee

        if transaction.fee_metric == wallet_pb2.FeeMetric.Value('BYTE'):
            self._fee_metric = "bytes"
        elif transaction.fee_metric == wallet_pb2.FeeMetric.Value('VBYTE'):
            self._fee_metric = "vbytes"
        elif transaction.fee_metric == wallet_pb2.FeeMetric.Value('WEI'):
            self._fee_metric = "wei"
        else:
            raise ValueError("Unknown fee metric")
        
        self._sat_metadata = {}
        self._evm_metadata =  {}
        if network.SUPPORTS_EVM:
            self._evm_metadata['from'] = transaction.ethlike_transaction.txfrom
            self._evm_metadata['to'] = transaction.ethlike_transaction.txto
            self._evm_metadata['amount'] = transaction.ethlike_transaction.amount
            self._evm_metadata['gasUsed'] = transaction.ethlike_transaction.gas
            self._evm_metadata['data'] = transaction.ethlike_transaction.data
        else:
            self._sat_metadata['feeRate'] = transaction.btclike_transaction.fee
            self._sat_metadata['inputs'] = []
            self._sat_metadata['outputs'] = []
            for i in transaction.btclike_transaction.inputs:
                i_ = {}
                i_['txid'] = i.txid
                i_['index'] = i.index
                i_['amount'] = i.amount
                i_['witness'] = []
                for w in i.witness_data:
                    i_['witness'].append(w)
                self._sat_metadata['inputs'].append(i_)
            for o in transaction.btclike_transaction.outputs:
                o_ = {}
                o_['address'] = o.address
                o_['index'] = o.index
                o_['amount'] = o.amount
                o_['spent'] = o.spent
                self._sat_metadata['outputs'].append(o_)
        
    def network(self):
        return self._network

    def txid(self):
        return self._txid
    
    def timestamp(self):
        return self._timestamp
    
    def confirmed(self):
        return self._confirmed

    def height(self):
        return self._height
    
    def total_fee(self, in_standard_units=True):
        if in_standard_units:
            if self._network.SUPPORTS_EVM:
                fee = self._total_fee / 1e18
            else:
                fee = self._total_fee / 1e8
        else:
            fee = self._total_fee
        return (fee, self._fee_metric)
        
    def evm_from(self):
        if not self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        return self._evm_metadata['from']
    
    def evm_to(self):
        if not self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        return self._evm_metadata['to']
    
    def evm_amount(self, in_standard_units=True):
        if not self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        if in_standard_units:
            return self._evm_metadata['amount'] / 1e18
        else:
            return self._evm_metadata['amount']
    
    def evm_gas(self):
        if not self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        return self._evm_metadata['gasUsed'] # always in WEI

    def evm_data(self):
        if not self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        return self._evm_metadata['data']
    
    def sat_feerate(self):
        if self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        return self._sat_metadata['feeRate'] # always in sats per byte or vbyte
    
    def sat_inputs(self, include_witness=False):
        if self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        inputs = []
        for i in self._sat_metadata['inputs']:
            if not include_witness and 'witness' in i.keys():
                del i['witness']
            inputs.append(i)
        return inputs
        
    def sat_outputs(self, only_unspent=False):
        if self._network.SUPPORTS_EVM:
            raise ValueError("Blockchain does not support this property")
        outputs = []
        for o in self._sat_metadata['outputs']:
            if not only_unspent or not o['spent']:
                outputs.append(o)
        return outputs





#!/usr/bin/env python
# flake8: noqa: C0301

"""Tests for address transaction, balance, and UTXO fetcher."""

import unittest
from zpywallet.address.btc import BitcoinAddress
from zpywallet.address.btctest import BitcoinTestAddress
from zpywallet.address.dash import DashAddress
from zpywallet.address.doge import DogecoinAddress
from zpywallet.address.eth import EthereumAddress
from zpywallet.address.ltc import LitecoinAddress
from zpywallet.errors import NetworkException

class TestAddress(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_btc_address(self):
        """Test fetching Bitcoin addresses."""
        b = BitcoinAddress(['3KzZceAGsA7HRxFzgbZxVJMAV9TJa8o97V', 'bc1plytzh6jqwltfq6l0ujt5ucz9csrlff4rfnxwmy3tkepkeyj3y2gskcf48c'])
        try:
            b.sync()
            b.get_transaction_history()
            b.get_utxos()
            b.get_balance()
        except NetworkException as e:
            pass

    def test_001_btctest_address(self):
        """Test fetching Bitcoin testnet addressses"""
        b = BitcoinTestAddress(['2NDNwoqdNvJ2jkBD8B6VVxNntpuKR4tkTSz', 'tb1qtnl457a54cz5gq0zgyf4n2xt7n9uuqxzs8jwrp'])
        try:
            b.sync()
            b.get_transaction_history()
            b.get_utxos()
            b.get_balance()
        except NetworkException as e:
            pass

    def test_002_dash_address(self):
        """Test fetching Dash addressses"""
        b = DashAddress(['XbzLCqAv8rkYmky6uEsxibHRUbHZU2XCKg'])
        try:
            b.sync()
            b.get_transaction_history()
            b.get_utxos()
            b.get_balance()
        except NetworkException as e:
            pass

    def test_003_doge_address(self):
        """Test fetching Dogecoin addressses"""
        b = DogecoinAddress(['DEhEbNiKjyQXSDCZGVh3iZSEBdLT6seLKA'])
        try:
            b.sync()
            b.get_transaction_history()
            b.get_utxos()
            b.get_balance()
        except NetworkException as e:
            pass

    def test_004_eth_address(self):
        """Test fetching Ethereum addressses"""
        b = EthereumAddress(['0x383d4669f177182f2c8c90cecd291190ea04edad'])
        try:
            b.sync()
            # b.get_transaction_history()  # On EVM chains, this scans all blocks and takes forever.
            b.get_balance()
        except NetworkException as e:
            pass
    
    def test_003_ltc_address(self):
        """Test fetching Litecoin addressses"""
        b = LitecoinAddress(['ltc1q9pw48v23gq9d2lqcss8yaqeh7fqzu4wrt6m6nr'])
        try:
            b.sync()
            b.get_transaction_history()
            b.get_utxos()
            b.get_balance()
        except NetworkException as e:
            pass


import requests

from ...errors import NetworkException

async def broadcast_transaction_btctest_blockstream(raw_transaction_hex):
    api_url = "https://blockstream.info/testnet/api/tx"
    payload = {"tx": raw_transaction_hex}

    try:
        response = requests.post(api_url, data=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))

    if response.status_code >= 300:
        raise NetworkException("Failed to broadcast testnet transaction using Blocksream API: {}".format(response.text))



import requests
import time

from functools import reduce

from ...errors import NetworkException
from ...utils.utils import convert_to_utc_timestamp
from ...generated import wallet_pb2

def deduplicate(elements):
    return reduce(lambda re, x: re+[x] if x not in re else re, elements, [])

class BlockcypherAddress:
    """
    A class representing a Dogecoin address.

    This class allows you to retrieve the balance and transaction history of a Dogecoin address using the Blockcypher API.

    Args:
        address (str): The human-readable Dogecoin address.
        api_key (str): The API key for accessing the Blockcypher API.

    Attributes:
        address (str): The human-readable Dogecoin address.
        api_key (str): The API key for accessing the Blockcypher API.

    Methods:
        get_balance(): Retrieves the balance of the Dogecoin address.
        get_transaction_history(): Retrieves the transaction history of the Dogecoin address.

    Raises:
        Exception: If the API request fails or the address balance/transaction history cannot be retrieved.
    """

    def _clean_tx(self, element):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['hash']
        
        if 'block_height' not in element.keys():
            new_element.confirmed = False
        elif element['block_height'] == -1:
            new_element.confirmed = False
        elif element['block_index'] == 0:
            new_element.confirmed = True
            new_element.height = 0
        else:
            new_element.confirmed = True
            new_element.height = element['block_height']
        
        if 'confirmed' in element.keys():
            new_element.timestamp = convert_to_utc_timestamp(element['confirmed'].split(".")[0].split('Z')[0], '%Y-%m-%dT%H:%M:%S')
        
        for vin in element['inputs']:
            txinput = new_element.btclike_transaction.inputs.add()
            txinput.txid = '' if 'prev_hash' not in vin.keys() else vin['prev_hash']
            txinput.index = vin['output_index']
            txinput.amount = 0 if 'output_value' not in vin.keys() else int(vin['output_value'])
        
        i = 0
        for vout in element['outputs']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(vout['value'])
            txoutput.index = i
            i += 1
            if vout['addresses']:
                txoutput.address = vout['addresses'][0]
            txoutput.spent = 'spent_by' in vout.keys()
        
        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])
        new_element.total_fee = total_inputs - total_outputs

        #size_element = element['vsize'] if 'vsize' in element.keys() else element['size']
        size_element = element['size']
        new_element.btclike_transaction.fee = int(new_element.total_fee // size_element)
        new_element.fee_metric = wallet_pb2.BYTE
        
        return new_element

    def __init__(self, addresses, request_interval=(3,1), transactions=None, api_key=None):
        """
        Initializes an instance of the BlockcypherAddress class.

        Args:
            addresses (list): A list of human-readable Dogecoin addresses.
            api_key (str): The API key for accessing the Blockcypher API.
            request_interval (tuple): A pair of integers indicating the number of requests allowed during
                a particular amount of seconds. Set to (0,N) for no rate limiting, where N>0.
        """
        self.addresses = addresses
        self.api_key = api_key
        self.requests, self.interval_sec = request_interval
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.transactions = deduplicate([*self._get_transaction_history()])
        self.height = self.get_block_height()

    def get_balance(self):
        """
        Retrieves the balance of the Dogecoin address.

        Returns:
            float: The balance of the Dogecoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos

    def get_block_height(self):
        """Returns the current block height."""

        url = "https://api.blockcypher.com/v1/doge/main"
        params = None
        if self.api_key:
            params = {"token", self.api_key}
        for attempt in range(3, -1, -1):
            if attempt == 0:
                raise NetworkException("Network request failure")
            try:
                response = requests.get(url, params=params, timeout=60)
                if response.status_code == 200:
                    data = response.json()
                else:
                    raise NetworkException("Failed to retrieve block height")
                break
            except requests.RequestException:
                pass
            except requests.exceptions.JSONDecodeError:
                pass

        self.height = data["height"]
        return self.height

    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Dogecoin address from cached data augmented with network data.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs
                    
        self.transactions = deduplicate(self.transactions)
        return self.transactions

    def _get_transaction_history(self, txhash=None):
        """
        Retrieves the transaction history of the Dogecoin address. (internal method that makes the network query)

        Parameters:
            txhash (str): Get all transactions before (and not including) txhash.
                Defaults to None, which disables this behavior.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        params = None
        if self.api_key:
            params = {"token", self.api_key}
        for address in self.addresses:
            interval = 50
            block_height = 0

            # Set a very high UTXO limit for those rare address that have crazy high input/output counts.
            txlimit = 10000

            url = f"https://api.blockcypher.com/v1/doge/main/addrs/{address}/full?limit={interval}&txlimit={txlimit}"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, params=params, timeout=60)
                    if response.status_code == 200:
                        data = response.json()
                        break
                    else:
                        raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass

            for tx in data["txs"]:
                time.sleep(self.interval_sec/(self.requests*len(data["txs"])))
                if txhash and tx["hash"] == txhash:
                    return
                yield self._clean_tx(tx)
            if 'hasMore' not in data.keys():
                return
            else:
                block_height = data["txs"][-1]["block_height"]
            
            while 'hasMore' in data.keys() and data['hasMore']:
                url = f"https://api.blockcypher.com/v1/ddogeash/main/addrs/{address}/full?limit={interval}&before={block_height}&txlimit={txlimit}"
                for attempt in range(3, -1, -1):
                    if attempt == 0:
                        raise NetworkException("Network request failure")
                    try:
                        response = requests.get(url, params=params, timeout=60)
                        break
                    except requests.RequestException:
                        pass

                if response.status_code == 200:
                    data = response.json()
                    for tx in data["txs"]:
                        time.sleep(self.interval_sec/(self.requests*len(data["txs"])))
                        if txhash and tx["hash"] == txhash:
                            return
                        yield self._clean_tx(tx)
                    if 'hasMore' not in data.keys():
                        return
                    else:
                        block_height = data["txs"][-1]["block_height"]
                else:
                    raise NetworkException("Failed to retrieve transaction history")



# Copyright (C) 2018-2022 The python-bitcoin-utils developers
#
# This file is part of python-bitcoin-utils
#
# It is subject to the license terms in the LICENSE file found in the top-level
# directory of this distribution.
#
# No part of python-bitcoin-utils, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE file.

import struct
import copy
import hashlib
from binascii import unhexlify, hexlify

from ..utils.utils import hash160

from ..network import BitcoinSegwitMainNet

from ..utils.base58 import b58encode_check

from ..utils.bech32 import bech32_encode

def p2sh_address(redeem_script, network=BitcoinSegwitMainNet):
    """Generate a P2SH (3) address from a redeem script."""
    return b58encode_check(bytes([network.SCRIPT_ADDRESS]) + bytes.fromhex(redeem_script))

def to_bytes(string, unhex=True):
    '''
	Converts a hex string to bytes
    '''
    if not string:
        return b''
    if unhex:
        try:
            if isinstance(string, bytes):
                string = string.decode()
            s = bytes.fromhex(string)
            return s
        except (TypeError, ValueError):
            pass
    if isinstance(string, bytes):
        return string
    else:
        return bytes(string, 'utf8')
    
def parse_varint(data):
    def bytes_to_int(b):
        return int.from_bytes(b, byteorder="little")
    varint_type = data[0]
    if varint_type < 0xfd:
        return varint_type, 1
    elif varint_type == 0xfd:
        return bytes_to_int(data[1:3]), 3
    elif varint_type == 0xfe:
        return bytes_to_int(data[1:5]), 5
    elif varint_type == 0xff:
        return bytes_to_int(data[1:9]), 9

# Bitcoin's op codes. Complete list at: https://en.bitcoin.it/wiki/Script
OP_CODES = {
    # constants
    'OP_0'                  : b'\x00',
    'OP_FALSE'              : b'\x00',
    'OP_PUSHDATA1'          : b'\x4c',
    'OP_PUSHDATA2'          : b'\x4d',
    'OP_PUSHDATA4'          : b'\x4e',
    'OP_1NEGATE'            : b'\x4f',
    'OP_1'                  : b'\x51',
    'OP_TRUE'               : b'\x51',
    'OP_2'                  : b'\x52',
    'OP_3'                  : b'\x53',
    'OP_4'                  : b'\x54',
    'OP_5'                  : b'\x55',
    'OP_6'                  : b'\x56',
    'OP_7'                  : b'\x57',
    'OP_8'                  : b'\x58',
    'OP_9'                  : b'\x59',
    'OP_10'                 : b'\x5a',
    'OP_11'                 : b'\x5b',
    'OP_12'                 : b'\x5c',
    'OP_13'                 : b'\x5d',
    'OP_14'                 : b'\x5e',
    'OP_15'                 : b'\x5f',
    'OP_16'                 : b'\x60',

    # flow control
    'OP_NOP'                : b'\x61',
    'OP_IF'                 : b'\x63',
    'OP_NOTIF'              : b'\x64',
    'OP_ELSE'               : b'\x67',
    'OP_ENDIF'              : b'\x68',
    'OP_VERIFY'             : b'\x69',
    'OP_RETURN'             : b'\x6a',

    # stack
    'OP_TOALTSTACK'         : b'\x6b',
    'OP_FROMALTSTACK'       : b'\x6c',
    'OP_IFDUP'              : b'\x73',
    'OP_DEPTH'              : b'\x74',
    'OP_DROP'               : b'\x75',
    'OP_DUP'                : b'\x76',
    'OP_NIP'                : b'\x77',
    'OP_OVER'               : b'\x78',
    'OP_PICK'               : b'\x79',
    'OP_ROLL'               : b'\x7a',
    'OP_ROT'                : b'\x7b',
    'OP_SWAP'               : b'\x7c',
    'OP_TUCK'               : b'\x7d',
    'OP_2DROP'              : b'\x6d',
    'OP_2DUP'               : b'\x6e',
    'OP_3DUP'               : b'\x6f',
    'OP_2OVER'              : b'\x70',
    'OP_2ROT'               : b'\x71',
    'OP_2SWAP'              : b'\x72',

    # splice
    #'OP_CAT'                : b'\x7e',
    #'OP_SUBSTR'             : b'\x7f',
    #'OP_LEFT'               : b'\x80',
    #'OP_RIGHT'              : b'\x81',
    'OP_SIZE'               : b'\x82',

    # bitwise logic
    #'OP_INVERT'             : b'\x83',
    #'OP_AND'                : b'\x84',
    #'OP_OR'                 : b'\x85',
    #'OP_XOR'                : b'\x86',
    'OP_EQUAL'              : b'\x87',
    'OP_EQUALVERIFY'        : b'\x88',

    # arithmetic
    'OP_1ADD'               : b'\x8b',
    'OP_1SUB'               : b'\x8c',
    #'OP_2MUL'               : b'\x8d',
    #'OP_2DIV'               : b'\x8e',
    'OP_NEGATE'             : b'\x8f',
    'OP_ABS'                : b'\x90',
    'OP_NOT'                : b'\x91',
    'OP_0NOTEQUAL'          : b'\x92',
    'OP_ADD'                : b'\x93',
    'OP_SUB'                : b'\x94',
    #'OP_MUL'                : b'\x95',
    #'OP_DIV'                : b'\x96',
    #'OP_MOD'                : b'\x97',
    #'OP_LSHIFT'             : b'\x98',
    #'OP_RSHIFT'             : b'\x99',
    'OP_BOOLAND'            : b'\x9a',
    'OP_BOOLOR'             : b'\x9b',
    'OP_NUMEQUAL'           : b'\x9c',
    'OP_NUMEQUALVERIFY'     : b'\x9d',
    'OP_NUMNOTEQUAL'        : b'\x9e',
    'OP_LESSTHAN'           : b'\x9f',
    'OP_GREATERTHAN'        : b'\xa0',
    'OP_LESSTHANOREQUAL'    : b'\xa1',
    'OP_GREATERTHANOREQUAL' : b'\xa2',
    'OP_MIN'                : b'\xa3',
    'OP_MAX'                : b'\xa4',
    'OP_WITHIN'             : b'\xa5',

    # crypto
    'OP_RIPEMD160'          : b'\xa6',
    'OP_SHA1'               : b'\xa7',
    'OP_SHA256'             : b'\xa8',
    'OP_HASH160'            : b'\xa9',
    'OP_HASH256'            : b'\xaa',
    'OP_CODESEPARATOR'      : b'\xab',
    'OP_CHECKSIG'           : b'\xac',
    'OP_CHECKSIGVERIFY'     : b'\xad',
    'OP_CHECKMULTISIG'      : b'\xae',
    'OP_CHECKMULTISIGVERIFY': b'\xaf',

    # locktime
    'OP_NOP2'               : b'\xb1',
    'OP_CHECKLOCKTIMEVERIFY': b'\xb1',
    'OP_NOP3'               : b'\xb2',
    'OP_CHECKSEQUENCEVERIFY': b'\xb2'
}

CODE_OPS = {
    # constants
    b'\x00':    'OP_0'                  , 
    b'\x4c':    'OP_PUSHDATA1'          , 
    b'\x4d':    'OP_PUSHDATA2'          , 
    b'\x4e':    'OP_PUSHDATA4'          , 
    b'\x4f':    'OP_1NEGATE'            , 
    b'\x51':    'OP_1'                  , 
    b'\x52':    'OP_2'                  , 
    b'\x53':    'OP_3'                  , 
    b'\x54':    'OP_4'                  , 
    b'\x55':    'OP_5'                  , 
    b'\x56':    'OP_6'                  , 
    b'\x57':    'OP_7'                  , 
    b'\x58':    'OP_8'                  , 
    b'\x59':    'OP_9'                  , 
    b'\x5a':    'OP_10'                 , 
    b'\x5b':    'OP_11'                 , 
    b'\x5c':    'OP_12'                 , 
    b'\x5d':    'OP_13'                 , 
    b'\x5e':    'OP_14'                 , 
    b'\x5f':    'OP_15'                 , 
    b'\x60':    'OP_16'                 , 

    # flow control
    b'\x61':    'OP_NOP'                , 
    b'\x63':    'OP_IF'                 , 
    b'\x64':    'OP_NOTIF'              , 
    b'\x67':    'OP_ELSE'               , 
    b'\x68':    'OP_ENDIF'              , 
    b'\x69':    'OP_VERIFY'             , 
    b'\x6a':    'OP_RETURN'             , 

    # stack
    b'\x6b':    'OP_TOALTSTACK'         , 
    b'\x6c':    'OP_FROMALTSTACK'       , 
    b'\x73':    'OP_IFDUP'              , 
    b'\x74':    'OP_DEPTH'              , 
    b'\x75':    'OP_DROP'               , 
    b'\x76':    'OP_DUP'                , 
    b'\x77':    'OP_NIP'                , 
    b'\x78':    'OP_OVER'               , 
    b'\x79':    'OP_PICK'               , 
    b'\x7a':    'OP_ROLL'               , 
    b'\x7b':    'OP_ROT'                , 
    b'\x7c':    'OP_SWAP'               , 
    b'\x7d':    'OP_TUCK'               , 
    b'\x6d':    'OP_2DROP'              , 
    b'\x6e':    'OP_2DUP'               , 
    b'\x6f':    'OP_3DUP'               , 
    b'\x70':    'OP_2OVER'              , 
    b'\x71':    'OP_2ROT'               , 
    b'\x72':    'OP_2SWAP'              , 

    # splice
    b'\x82':    'OP_SIZE'               , 

    # bitwise logic
    b'\x87':    'OP_EQUAL'              , 
    b'\x88':    'OP_EQUALVERIFY'        , 

    # arithmetic
    b'\x8b':    'OP_1ADD'               , 
    b'\x8c':    'OP_1SUB'               , 
    b'\x8f':    'OP_NEGATE'             , 
    b'\x90':    'OP_ABS'                , 
    b'\x91':    'OP_NOT'                , 
    b'\x92':    'OP_0NOTEQUAL'          , 
    b'\x93':    'OP_ADD'                , 
    b'\x94':    'OP_SUB'                , 
    b'\x9a':    'OP_BOOLAND'            , 
    b'\x9b':    'OP_BOOLOR'             , 
    b'\x9c':    'OP_NUMEQUAL'           , 
    b'\x9d':    'OP_NUMEQUALVERIFY'     , 
    b'\x9e':    'OP_NUMNOTEQUAL'        , 
    b'\x9f':    'OP_LESSTHAN'           , 
    b'\xa0':    'OP_GREATERTHAN'        , 
    b'\xa1':    'OP_LESSTHANOREQUAL'    , 
    b'\xa2':    'OP_GREATERTHANOREQUAL' , 
    b'\xa3':    'OP_MIN'                , 
    b'\xa4':    'OP_MAX'                , 
    b'\xa5':    'OP_WITHIN'             , 

    # crypto
    b'\xa6':    'OP_RIPEMD160'          , 
    b'\xa7':    'OP_SHA1'               , 
    b'\xa8':    'OP_SHA256'             , 
    b'\xa9':    'OP_HASH160'            , 
    b'\xaa':    'OP_HASH256'            , 
    b'\xab':    'OP_CODESEPARATOR'      , 
    b'\xac':    'OP_CHECKSIG'           , 
    b'\xad':    'OP_CHECKSIGVERIFY'     , 
    b'\xae':    'OP_CHECKMULTISIG'      , 
    b'\xaf':    'OP_CHECKMULTISIGVERIFY', 

    # locktime
    #b'\xb1':    'OP_NOP2'               , 
    b'\xb1':    'OP_CHECKLOCKTIMEVERIFY', 
    #b'\xb2':    'OP_NOP3'               , 
    b'\xb2':    'OP_CHECKSEQUENCEVERIFY' 
}

class Script:
    """Represents any script in Bitcoin

    A Script contains just a list of OP_CODES and also knows how to serialize
    into bytes

    Attributes
    ----------
    script : list
        the list with all the script OP_CODES and data

    Methods
    -------
    to_bytes()
        returns a serialized byte version of the script

    to_hex()
        returns a serialized version of the script in hex

    get_script()
        returns the list of strings that makes up this script

    copy()
        creates a copy of the object (classmethod)

    Raises
    ------
    ValueError
        If string data is too large or integer is negative
    """

    def __init__(self, script, network=BitcoinSegwitMainNet):
        """See Script description"""

        self.script = script
        self.network = network


    @classmethod
    def copy(cls, script):
        """Deep copy of Script"""
        scripts = copy.deepcopy(script.script)
        return cls(scripts)

    def __str__(self):
        return str(self.script)

    def __repr__(self):
        return self.__str__()

    def _op_push_data(self, data):
        """Converts data to appropriate OP_PUSHDATA OP code including length

        0x01-0x4b           -> just length plus data bytes
        0x4c-0xff           -> OP_PUSHDATA1 plus 1-byte-length plus data bytes
        0x0100-0xffff       -> OP_PUSHDATA2 plus 2-byte-length plus data bytes
        0x010000-0xffffffff -> OP_PUSHDATA4 plus 4-byte-length plus data bytes

        Also note that according to standarardness rules (BIP-62) the minimum
        possible PUSHDATA operator must be used!
        """

        # expects data in hexadecimal characters and converts appropriately
        data_bytes = unhexlify(data)

        if len(data_bytes) < 0x4c:
            return chr(len(data_bytes)).encode() + data_bytes
        elif len(data_bytes) < 0xff:
            return b'\x4c' + chr(len(data_bytes)).encode() + data_bytes
        elif len(data_bytes) < 0xffff:
            return b'\x4d' + struct.pack('<H', len(data_bytes)) + data_bytes
        elif len(data_bytes) < 0xffffffff:
            return b'\x4e' + struct.pack('<I', len(data_bytes)) + data_bytes
        else:
            raise ValueError("Data too large. Cannot push into script")


    def _segwit_op_push_data(self, data):
        # expects data in hexadecimal characters and converts to bytes with
        # varint (or compact size) length prefix.
        data_bytes = unhexlify(data)

        # return prepended varint (compact size) length to data bytes
        return parse_varint(data_bytes)[0] + data_bytes



    def _push_integer(self, integer):
        """Converts integer to bytes; as signed little-endian integer

        Currently supports only positive integers
        """

        if integer < 0:
            raise ValueError('Integer is currently required to be positive.')

        # bytes required to represent the integer
        number_of_bytes = (integer.bit_length() + 7) // 8

        # convert to little-endian bytes
        integer_bytes = integer.to_bytes(number_of_bytes, byteorder='little')

        # if last bit is set then we need to add sign to signify positive
        # integer
        if integer & (1 << number_of_bytes*8 - 1):
            integer_bytes += b'\x00'

        return self._op_push_data( hexlify(integer_bytes) )


    def to_bytes(self):
        """Converts the script to bytes

        If an OP code the appropriate byte is included according to:
        https://en.bitcoin.it/wiki/Script
        If not consider it data (signature, public key, public key hash, etc.) and
        and include with appropriate OP_PUSHDATA OP code plus length
        """
        script_bytes = b''
        for token in self.script:
            # add op codes directly
            if token in OP_CODES:
                script_bytes += OP_CODES[token]
            # if integer between 0 and 16 add the appropriate op code
            elif type(token) is int and token >= 0 and token <= 16:
                script_bytes += OP_CODES['OP_' + str(token)]
            # it is data, so add accordingly
            else:
                if type(token) is int:
                    script_bytes += self._push_integer(token)
                else:
                    script_bytes += self._op_push_data(token)

        return script_bytes

    @staticmethod
    def from_raw(scriptraw, has_segwit=False, network=BitcoinSegwitMainNet):
        """
        Imports a Script commands list from raw hexadecimal data
            Attributes
            ----------
            txinputraw : string (hex)
                The hexadecimal raw string representing the Script commands
            has_segwit : boolean
                Is the Tx Input segwit or not
        """
        scriptraw = to_bytes(scriptraw)
        commands = []
        index = 0
        while index < len(scriptraw):
            byte = scriptraw[index]
            if bytes([byte]) in CODE_OPS:
                commands.append(CODE_OPS[bytes([byte])])
                index = index + 1
                #handle the 3 special bytes 0x4c,0x4d,0x4e if the transaction is not segwit type
            elif has_segwit == False and bytes([byte]) == b'\x4c':
                bytes_to_read = int.from_bytes(scriptraw[index + 1], "little")
                index = index + 1
                commands.append(scriptraw[index: index + bytes_to_read].hex())
                index = index + bytes_to_read
            elif has_segwit == False and bytes([byte]) == b'\x4d':
                bytes_to_read = int.from_bytes(scriptraw[index:index + 2], "little")
                index = index + 2
                commands.append(scriptraw[index: index + bytes_to_read].hex())
                index = index + bytes_to_read
            elif has_segwit == False and bytes([byte]) == b'\x4e':
                bytes_to_read = int.from_bytes(scriptraw[index:index + 4], "little")
                index = index + 4
                commands.append(scriptraw[index: index + bytes_to_read].hex())
                index = index + bytes_to_read
            else:
                data_size, size = parse_varint(scriptraw[index:index + 9])
                commands.append(scriptraw[index + size:index + size + data_size].hex())
                index = index + data_size + size


        return Script(script=commands, network=network)


    def to_hex(self):
        """Converts the script to hexadecimal"""

        b = self.to_bytes()
        return hexlify(b).decode('utf-8')


    def get_script(self):
        """Returns script as array of strings"""
        return self.script

    def is_p2pkh(self):
        """Checks whether the transaction is P2PKH"""
        try:
            return self.script[0] == "OP_DUP" and \
            self.script[1] == "OP_HASH160" and \
            self.script[3] == "OP_EQUALVERIFY" and \
            self.script[4] == "OP_CHECKSIG" and \
            len(self.script) == 5
        except KeyError:
            return False

    def is_p2sh(self):
        """Checks whether the transaction is P2SH"""
        try:
            return self.script[0] == "OP_HASH160" and \
            self.script[2] == "OP_EQUAL" and \
            len(self.script) == 3
        except KeyError:
            return False

    def is_p2wpkh(self):
        """Checks whether the transaction is P2WPKH"""
        # The script is in hex
        try:
            return self.script[0] == "OP_0" and \
            len(self.script[1]) == 20*2 and \
            len(self.script) == 2
        except KeyError:
            return False
        
    def is_p2wsh(self):
        """Checks whether the transaction is P2WSH"""
        # The script is in hex
        try:
            return self.script[0] == "OP_0" and \
            len(self.script[1]) == 32*2 and \
            len(self.script) == 2
        except KeyError:
            return False
        

    def is_p2tr(self):
        """Checks whether the transaction is P2TR"""
        # The script is in hex
        try:
            return self.script[0] == "OP_1" and \
            len(self.script[1]) == 32*2 and \
            len(self.script) == 2
        except KeyError:
            return False
        
    def to_p2pkh(self):
        """Creates the P2PKH address from the script."""
        if self.is_p2pkh():
            return b58encode_check(bytes([self.network.PUBKEY_ADDRESS]) + bytes.fromhex(self.script[2]))
        else:
            return None
    
    def to_p2sh(self):
        """Creates the P2PKH address from the script."""
        if self.is_p2sh():
            return b58encode_check(bytes([self.network.SCRIPT_ADDRESS]) + bytes.fromhex(self.script[1]))
        else:
            return None
        
    def to_p2wpkh(self):
        """Creates the P2WPKH address from the script."""
        if self.is_p2wpkh():
            return bech32_encode(self.network.BECH32_PREFIX, 0, bytes.fromhex(self.script[1]))
        else:
            return None
    
    def to_p2wsh(self):
        """Creates the P2WSH address from the script."""
        if self.is_p2wsh():
            return bech32_encode(self.network.BECH32_PREFIX, 0, bytes.fromhex(self.script[1]))
        else:
            return None

    def to_p2tr(self):
        """Creates the P2TR address from the script."""
        if self.is_p2tr():
            return bech32_encode(self.network.BECH32_PREFIX, 1, bytes.fromhex(self.script[1]))
        else:
            return None


    def to_p2sh_script_pub_key(self):
        """Converts script to p2sh scriptPubKey (locking script)

        Calculates the hash160 (via the address) of the script and uses it to
        construct a P2SH script.
        """

        redeem_script = hash160(self.to_hex()) # script hash
        #address = p2sh_address(redeem_script)
        return Script(['OP_HASH160', redeem_script, 'OP_EQUAL'])

    def to_p2wsh_script_pub_key(self):
        """Converts script to p2wsh scriptPubKey (locking script)

        Calculates the sha256 of the script and uses it to construct a P2WSH script.
        """
        sha256 = hashlib.sha256( self.to_bytes() ).digest()
        return Script(['OP_0', hexlify(sha256).decode('utf-8')])




__all__ = [
    'bitaps',
    'blockchain_info',
    'blockchair',
    'blockcypher',
    'blockstream',
    'coinify',
    'esplora',
    'f2pool',
    'fullnode',
    'mempool_space',
    'smartbit',
    'viabtc'
]

from .all import broadcast_transaction_btc, tx_hash_btc


import random
import requests
from ...errors import NetworkException

async def broadcast_transaction_doge_full_node(raw_transaction_hex, **kwargs):
    user = kwargs.get('user')
    password = kwargs.get('password')
    url = kwargs.get('url')
    if user and password:
        rpc_url = f"http://{user}:{password}@{url}"
    else:
        rpc_url = url
        
    payload = {
        "jsonrpc": "2.0",
        "id": f"{random.randint(1, (2<<31) - 1)}",
        "method": "sendrawtransaction",
        "params": [raw_transaction_hex],
    }

    try:
        response = requests.post(rpc_url, json=payload, timeout=30)
    except Exception as e:
        raise NetworkException(f"Failed to connect to RPC interface: {str(e)}")

    result = response.json()

    if "error" in result:
        raise NetworkException(f"Failed to broadcast Dogecoin transaction using full node: {result['error']}")
    
    return result["result"]



"""
The keccak module provides an implementation of the Keccak hash function, including various parameter presets
such as Keccak224, Keccak256, Keccak384, and Keccak512. The module offers a hashlib-compatible interface for
easy integration into existing codebases.
"""

from math import log
from operator import xor
from copy import deepcopy
from functools import reduce
from binascii import hexlify

# The round constants used in the Keccak-f permutation.
RoundConstants = [
  0x0000000000000001,   0x0000000000008082,   0x800000000000808A,   0x8000000080008000,
  0x000000000000808B,   0x0000000080000001,   0x8000000080008081,   0x8000000000008009,
  0x000000000000008A,   0x0000000000000088,   0x0000000080008009,   0x000000008000000A,
  0x000000008000808B,   0x800000000000008B,   0x8000000000008089,   0x8000000000008003,
  0x8000000000008002,   0x8000000000000080,   0x000000000000800A,   0x800000008000000A,
  0x8000000080008081,   0x8000000000008080,   0x0000000080000001,   0x8000000080008008
]

# The rotation constants used in the Keccak-f permutation.
RotationConstants = [
  [  0,  1, 62, 28, 27, ],
  [ 36, 44,  6, 55, 20, ],
  [  3, 10, 43, 25, 39, ],
  [ 41, 45, 15, 21,  8, ],
  [ 18,  2, 61, 56, 14, ]
]

# Bit masks used for circular rotations in the Keccak-f permutation.
Masks = [(1 << i) - 1 for i in range(65)]

def bits2bytes(x):
    """Converts the given number of bits to the corresponding number of bytes, rounding up if necessary."""
    return (int(x) + 7) // 8

def rol(value, left, bits):
    """
    Circularly rotate 'value' to the left,
    treating it as a quantity of the given size in bits.
    """
    top = value >> (bits - left)
    bot = (value & Masks[bits - left]) << left
    return bot | top

def ror(value, right, bits):
    """
    Circularly rotate 'value' to the right,
    treating it as a quantity of the given size in bits.
    """
    top = value >> right
    bot = (value & Masks[right]) << (bits - right)
    return bot | top

def multirate_padding(used_bytes, align_bytes):
    """
     Generates padding bytes according to the Keccak padding scheme,
     ensuring alignment to the specified number of bytes.
    """
    padlen = align_bytes - used_bytes
    if padlen == 0:
        padlen = align_bytes
    # note: padding done in 'internal bit ordering', wherein LSB is leftmost
    if padlen == 1:
        return [0x81]
    else:
        return [0x01] + ([0x00] * (padlen - 2)) + [0x80]

def keccak_f(state):
    """
    Performs the Keccak-f permutation on the given Keccak state. It applies multiple
    rounds of theta, rho, pi, chi, and iota operations to mutate the state. It operates
    on and mutates the passed-in KeccakState.  It returns nothing.
    """
    def f_round(A, RC):
        W, H = state.W, state.H
        rangeW, rangeH = state.rangeW, state.rangeH
        lanew = state.lanew
        zero = state.zero
    
        # theta
        C = [reduce(xor, A[x]) for x in rangeW]
        D = [0] * W
        for x in rangeW:
            D[x] = C[(x - 1) % W] ^ rol(C[(x + 1) % W], 1, lanew)
            for y in rangeH:
                A[x][y] ^= D[x]
        
        # rho and pi
        B = zero()
        for x in rangeW:
            for y in rangeH:
                B[y % W][(2 * x + 3 * y) % H] = rol(A[x][y], RotationConstants[y][x], lanew)
                
        # chi
        for x in rangeW:
            for y in rangeH:
                A[x][y] = B[x][y] ^ ((~ B[(x + 1) % W][y]) & B[(x + 2) % W][y])
        
        # iota
        A[0][0] ^= RC

    l = int(log(state.lanew, 2))
    nr = 12 + 2 * l
    
    for ir in range(nr):
        f_round(state.s, RoundConstants[ir])

class KeccakState(object):
    """
    A keccak state container.
    
     Represents the internal state of the Keccak algorithm.
     It maintains the state as a 5x5 table of integers and provides methods for
     manipulating the state, converting between byte sequences and lanes, and
     formatting the state as a hexadecimal string.
    """
    W = 5
    H = 5
    
    rangeW = list(range(W))
    rangeH = list(range(H))
    
    @staticmethod
    def zero():
        """
        Returns an zero state table.
        """
        return [[0] * KeccakState.W for x in KeccakState.rangeH]
    
    @staticmethod
    def format(st):
        """
        Formats the given state as hex, in natural byte order.
        """
        rows = []
        def fmt(x): return '%016x' % x
        for y in KeccakState.rangeH:
            row = []
            for x in KeccakState.rangeW:
                row.append(fmt(st[x][y]))
            rows.append(' '.join(row))
        return '\n'.join(rows)
    
    @staticmethod
    def lane2bytes(s, w):
        """
        Converts the lane s to a sequence of byte values,
        assuming a lane is w bits.
        """
        o = []
        for b in range(0, w, 8):
            o.append((s >> b) & 0xff)
        return o
    
    @staticmethod
    def bytes2lane(bb):
        """
        Converts a sequence of byte values to a lane.
        """
        r = 0
        for b in reversed(bb):
            r = r << 8 | b
        return r
    
    def __init__(self, bitrate, b):
        self.bitrate = bitrate
        self.b = b
        
        # only byte-aligned
        assert self.bitrate % 8 == 0
        self.bitrate_bytes = bits2bytes(self.bitrate)
        
        assert self.b % 25 == 0
        self.lanew = self.b // 25
        
        self.s = KeccakState.zero()
    
    def __str__(self):
        return KeccakState.format(self.s)
    
    def absorb(self, bb):
        """
        Mixes in the given bitrate-length string to the state.
        """
        assert len(bb) == self.bitrate_bytes
        
        bb += [0] * bits2bytes(self.b - self.bitrate)
        i = 0
        
        for y in self.rangeH:
            for x in self.rangeW:
                self.s[x][y] ^= KeccakState.bytes2lane(bb[i:i + 8])
                i += 8
    
    def squeeze(self):
        """
        Returns the bitrate-length prefix of the state to be output.
        """
        return self.get_bytes()[:self.bitrate_bytes]
    
    def get_bytes(self):
        """
        Convert whole state to a byte string.
        """
        out = [0] * bits2bytes(self.b)
        i = 0
        for y in self.rangeH:
            for x in self.rangeW:
                v = KeccakState.lane2bytes(self.s[x][y], self.lanew)
                out[i:i+8] = v
                i += 8
        return out
    
    def set_bytes(self, bb):
        """
        Set whole state from byte string, which is assumed
        to be the correct length.
        """
        i = 0
        for y in self.rangeH:
            for x in self.rangeW:
                self.s[x][y] = KeccakState.bytes2lane(bb[i:i+8])
                i += 8

class KeccakSponge(object):
    """
    Implements the sponge construction of the Keccak algorithm. It absorbs
    input data, applies the Keccak-f permutation, and produces output data
    based on the specified bitrate and capacity.
    """
    def __init__(self, bitrate, width, padfn, permfn):
        self.state = KeccakState(bitrate, width)
        self.padfn = padfn
        self.permfn = permfn
        self.buffer = []
        
    def copy(self):
        """ Creates a deep copy of the KeccakHash object, including the internal state. """
        return deepcopy(self)
        
    def absorb_block(self, bb):
        """ Absorbs a block of data of the bitrate length into the sponge's internal state. """
        assert len(bb) == self.state.bitrate_bytes
        self.state.absorb(bb)
        self.permfn(self.state)
    
    def absorb(self, s):
        """Absorbs input strings or bytes as byte data. """
        if type(s) is str:
            self.buffer += bytes(s, 'latin1')
        elif type(s) is bytes:
            self.buffer += s
        else:
            raise TypeError("Expected bytes or str")
        
        while len(self.buffer) >= self.state.bitrate_bytes:
            self.absorb_block(self.buffer[:self.state.bitrate_bytes])
            self.buffer = self.buffer[self.state.bitrate_bytes:]
    
    def absorb_final(self):
        """
        Used to apply padding to the remaining input data and absorb it into
        the sponge's internal state.
        """
        padded = self.buffer + self.padfn(len(self.buffer), self.state.bitrate_bytes)
        self.absorb_block(padded)
        self.buffer = []
        
    def squeeze_once(self):
        """Generates a single block of squeezed output data from the sponge's internal state."""
        rc = self.state.squeeze()
        self.permfn(self.state)
        return rc
    
    def squeeze(self, l):
        """Generates squeezed output data of the specified length from the sponge's internal state.
        
        Args:
            l (int): Length of squeezed data to return."""
        Z = self.squeeze_once()
        while len(Z) < l:
            Z += self.squeeze_once()
        return Z[:l]

class KeccakHash(object):
    """
    The Keccak hash function, with a hashlib-compatible interface.

    Represents a Keccak hash object with customizable bitrate, capacity, and output length.
    It provides methods for updating the hash state with input data and generating the
    final hash value.
    """
    def __init__(self, bitrate_bits, capacity_bits, output_bits):
        # our in-absorption sponge. this is never given padding
        assert bitrate_bits + capacity_bits in (25, 50, 100, 200, 400, 800, 1600)
        self.sponge = KeccakSponge(bitrate_bits, bitrate_bits + capacity_bits,
                                   multirate_padding,
                                   keccak_f)
        
        # hashlib interface members
        assert output_bits % 8 == 0
        self.digest_size = bits2bytes(output_bits)
        self.block_size = bits2bytes(bitrate_bits)
    
    def __repr__(self):
        inf = (self.sponge.state.bitrate,
               self.sponge.state.b - self.sponge.state.bitrate,
               self.digest_size * 8)
        return '<KeccakHash with r=%d, c=%d, image=%d>' % inf
    
    def copy(self):
        """ Creates a deep copy of the KeccakHash object, including the internal state. """
        return deepcopy(self)
    
    def update(self, s):
        """
        Updates the hash state by absorbing the input data `s`. The input data can be either a string or bytes object.

        Args:
            s (str or bytes): The input data to update the hash state with.
        """
        self.sponge.absorb(s)
    
    def digest(self):
        """
        Retrieves the final hash value as a bytes object. The hash state is finalized before generating the digest.

        Returns:
            bytes: The final hash value as a bytes object.
        """
        finalised = self.sponge.copy()
        finalised.absorb_final()
        digest = bytes(finalised.squeeze(self.digest_size))
        return digest
    
    def hexdigest(self):
        """
        Retrieves the final hash value as a hex string. The hash state is finalized before generating the digest.

        Returns:
            str: The final hash value as a hex string.
        """
        return hexlify(self.digest()).decode()
    
    @staticmethod
    def preset(bitrate_bits, capacity_bits, output_bits):
        """
        Returns a factory function for the given bitrate, sponge capacity and output length.
        The function accepts an optional initial input, ala hashlib.
        """
        def create(initial_input = None):
            h = KeccakHash(bitrate_bits, capacity_bits, output_bits)
            if initial_input is not None:
                h.update(initial_input)
            return h
        return create

# SHA3 parameter presets
# Keccak224: Creates a KeccakHash object with a bitrate of 1152 bits, a capacity of 448 bits, and an output length of 224 bits.
# Keccak256: Creates a KeccakHash object with a bitrate of 1088 bits, a capacity of 512 bits, and an output length of 256 bits.
# Keccak384: Creates a KeccakHash object with a bitrate of 832 bits, a capacity of 768 bits, and an output length of 384 bits.
# Keccak512: Creates a KeccakHash object with a bitrate of 576 bits, a capacity of 1024 bits, and an output length of 512 bits.
Keccak224 = KeccakHash.preset(1152, 448, 224)
Keccak256 = KeccakHash.preset(1088, 512, 256)
Keccak384 = KeccakHash.preset(832, 768, 384)
Keccak512 = KeccakHash.preset(576, 1024, 512)

def to_checksum_address(address):
    address = address.lower().replace('0x', '')
    keccak_hash = Keccak256(address.encode('utf-8')).hexdigest()
    checksum_address = '0x'
    
    for i in range(len(address)):
        if int(keccak_hash[i], 16) >= 8:
            checksum_address += address[i].upper()
        else:
            checksum_address += address[i]
    return checksum_address

def is_checksum_address(address):
    address = address.replace('0x', '')
    address_hash = Keccak256(address.lower().encode('utf-8')).hexdigest()
    
    for i in range(0, 40):
        # The nth letter should be uppercase if the nth digit of casemap is 1
        if ((int(address_hash[i], 16) > 7 and address[i].upper() != address[i]) or
                (int(address_hash[i], 16) <= 7 and address[i].lower() != address[i])):
            return False
    return True